__all__ = ['OMC']

import aesara_theano_fallback.tensor as T
from astropy.stats import mad_std
from celerite2.theano import terms as GPterms
from celerite2.theano import GaussianProcess
import numpy as np
import pymc3 as pm
import pymc3_ext as pmx
from scipy import stats
from scipy.ndimage import uniform_filter, median_filter
from sklearn.cluster import KMeans
from alderaan.constants import pi

class OMC:
    def __init__(self, ephemeris):
        # set initial O-C estimates
        self.index = ephemeris.index
        self.xtime = ephemeris.ttime
        self.yobs = ephemeris.ttime - ephemeris.eval_linear_ephemeris()
        self.yerr = ephemeris.error
        self.ymod = None

        # flag outliers
        self.quality = self.quick_flag_outliers()
        self.out_prob = None

        # set static reference period, epoch, and linear ephemeris
        self._set_static_references(ephemeris)

        # set nominal peak frequency and false alarm probability
        self.peakfreq = None
        self.peakfap = None


    def _set_static_references(self, ephemeris):
        self._static_period = ephemeris._static_period.copy()
        self._static_epoch = ephemeris._static_epoch.copy()


    def quick_flag_outliers(self, sigma_cut=5.0):
        if len(self.yobs) > 16:
            ysmooth = uniform_filter(median_filter(self.yobs, size=5, mode='mirror'), size=5)
        else:
            ysmooth = np.median(self.yobs)

        if len(self.yobs) > 4:
            quality = np.abs(self.yobs - ysmooth)/mad_std(self.yobs - ysmooth) < sigma_cut
        else:
            quality = np.zeros(len(self.yobs), dtype=bool)

        return quality


    def poly_model(self, polyorder, ignore_bad=True, xt_predict=None):
        """
        Build a PyMC3 model to fit TTV observed-minus-calculated data
        Fits data with a polynomial

        Arguments
        ----------
        polyorder : int
            polynomial order
        ignore_bad : bool
            True (default) to exclude bad quality times from model
        xt_predict : ndarray
            (optional) time values to predict OMC model

        Returns
        -------
        model : pm.Model()
        """
        # quality mask
        if ignore_bad:
            q = self.quality
        else:
            q = np.ones(len(self.xtime), dtype=bool)

        # times where trend will be predicted
        if xt_predict is None:
            xt_predict = self.xtime

        # create Vandermonde matrices
        Xt = np.vander(self.xtime[q], N=polyorder+1, increasing=True)
        Xp = np.vander(xt_predict, N=polyorder+1, increasing=True)
            
        # build pymc model
        with pm.Model() as model:
            C = pm.Normal("C", mu=0, sd=10, shape=polyorder+1)

            trend = pm.Deterministic("trend", pm.math.dot(Xt,C))
            pred = pm.Deterministic("pred", pm.math.dot(Xp,C))

            pm.Normal("obs", mu=trend, sd=self.yerr[q], observed=self.yobs[q])

        return model


    def sin_model(self, ttv_period=None, ignore_bad=True, xt_predict=None):
        """
        Build a PyMC3 model to fit TTV observed-minus-calculated data
        Fits data with a single-frequency sinusoid

        Arguments
        ----------
        ttv_period : float
            pre-estimated sinusoid period; the model places tight priors on this period
        ignore_bad : bool
            True (default) to exclude bad quality times from model
        xt_predict : ndarray
            (optional) time values to predict OMC model

        Returns
        -------
        model : pm.Model()
        """
        # quality mask
        if ignore_bad:
            q = self.quality
        else:
            q = np.ones(len(self.xtime), dtype=bool)

        # times where trend will be predicted
        if xt_predict is None:
            xt_predict = self.xtime

        # nominal TTV period
        if (ttv_period is None) & (self.peakfreq is None):
            ttv_period = 2 * (self.xtime[q].max() - self.xtime[q].min())
        elif (ttv_period is None) & (self.peakfreq is not None):
            ttv_period = 1 / self.peakfreq
        
        # convenience function
        def _sin_fxn(A, B, f, xt):
                return A * T.sin(2 * pi * f * xt) + B * T.cos(2 * pi * f * xt)

        # build pymc model
        with pm.Model() as model:
            df = 1 / (self.xtime[q].max() - self.xtime[q].min())
            f = pm.Normal("f", mu=1 / ttv_period, sd=df)
            Ah = pm.Normal("Ah", mu=0, sd=5 * np.std(self.yobs[q]))
            Bk = pm.Normal("Bk", mu=0, sd=5 * np.std(self.yobs[q]))

            trend = pm.Deterministic("trend", _sin_fxn(Ah, Bk, f, self.xtime[q]))
            pred = pm.Deterministic("pred", _sin_fxn(Ah, Bk, f, xt_predict))

            pm.Normal("obs", mu=trend, sd=self.yerr[q], observed=self.yobs[q])

        return model


    def matern32_model(self, ignore_bad=True, xt_predict=None):
        """
        Build a PyMC3 model to fit TTV observed-minus-calculated data
        Fits data with a regularized Matern-3/2 GP kernel

        Arguments
        ----------
        ignore_bad : bool
            True (default) to exclude bad quality times from model
        xt_predict : ndarray
            (optional) time values to predict OMC model

        Returns
        -------
        model : pm.Model()
        """
        # quality mask
        if ignore_bad:
            q = self.quality
        else:
            q = np.ones(len(self.xtime), dtype=bool)

        # times where trend will be predicted
        if xt_predict is None:
            xt_predict = self.xtime

        # delta between each transit time
        dx = np.mean(np.diff(self.xtime))

        # maximum GP amplitude
        ymax = np.sqrt(mad_std(self.yobs) ** 2 - np.median(self.yerr) ** 2)

        # pymc model
        with pm.Model() as model:
            # build the kernel
            log_sigma = pm.Bound(pm.Normal, upper=np.log(ymax))(
                "log_sigma", mu=np.log(ymax), sd=5
            )

            rho = pm.Uniform("rho", lower=2 * dx, upper=self.xtime[q].max() - self.xtime[q].min())
            kernel = GPterms.Matern32Term(sigma=T.exp(log_sigma), rho=rho)
            mean = pm.Normal("mean", mu=np.mean(self.yobs[q]), sd=np.std(self.yobs[q]))
            
            # define the GP and factorize the covariance matrix
            gp = GaussianProcess(kernel, t=self.xtime[q], yerr=self.yerr[q], mean=mean)
            gp.compute(self.xtime[q], yerr=self.yerr[q])

            # track GP prediction, covariance, and degrees of freedom
            trend, cov = gp.predict(self.yobs[q], self.xtime[q], return_cov=True)

            trend = pm.Deterministic("trend", trend)
            cov = pm.Deterministic("cov", cov)
            pred = pm.Deterministic("pred", gp.predict(self.yobs[q], xt_predict))
            dof = pm.Deterministic("dof", pm.math.trace(cov / self.yerr[q]**2))

            # add marginal likelihood to model
            lnlike = gp.log_likelihood(self.yobs[q])
            pm.Potential("lnlike", lnlike)

        return model


    def sample(self, model, progressbar=False):
        """
        Docstring
        """
        with model:
            map_soln = pmx.optimize(start=model.test_point, 
                                    progress=progressbar
                                    )
            
            trace = pmx.sample(tune=8000,
                               draws=2000,
                               chains=2,
                               target_accept=0.95,
                               start=map_soln,
                               progressbar=progressbar,
                               return_inferencedata=False
                              )
        return trace


    def identify_significant_frequencies(self, critical_fap):
        """
        Identify significant periodic frequencies using a Lomb-Scargle periodogram

        Arguments
        ---------
        critical_fap : float
            false alarm probability threshold to consider a frequency significant (default=0.1)
        """
        q = self.quality
        npts = np.sum(self.quality)

        peakfreq = None
        peakfap = None

        if npts > 8:
            try:
                xf, yf, freqs, faps = self.LS_estimator(self.xtime[q], self.yobs[q], fap=critical_fap)

                if len(freqs) > 0:
                    if freqs[0] > xf.min():
                        peakfreq = freqs[0]
                        peakfap = faps[0]

            except Exception:
                pass
        
        return peakfreq, peakfap
    

    @staticmethod
    def LS_estimator(x, y, fsamp=None, fap=0.1, return_levels=False, max_peaks=2):
        """
        Generate a Lomb-Scargle periodogram and identify significant frequencies
        * assumes that data are nearly evenly sampled
        * optimized for finding marginal periodic TTV signals in OMC data
        * may not perform well for other applications

        Arguments
        ---------
        x : array-like
            1D array of x data values; should be monotonically increasing
        y : array-like
            1D array of corresponding y data values, len(x)
        fsamp: float
            nominal sampling frequency; if not provided it will be calculated from the data
        fap : float
            false alarm probability threshold to consider a frequency significant (default=0.1)

        Returns
        -------
        xf : ndarray
            1D array of frequencies
        yf : ndarray
            1D array of corresponding response
        freqs : list
            signficant frequencies
        faps : list
            corresponding false alarm probabilities
        """
        # get sampling frequency
        if fsamp is None:
            fsamp = 1 / np.min(x[1:] - x[:-1])

        # Hann window to reduce ringing
        hann = signal.windows.hann(len(x))
        hann /= np.sum(hann)

        # identify any egregious outliers
        out = np.abs(y - np.median(y)) / mad_std(y - np.median(y)) > 5.0
        xt = x[~out]
        yt = y[~out]

        freqs = []
        faps = []

        loop = True
        while loop:
            lombscargle = LombScargle(xt, yt * hann[~out])
            xf, yf = lombscargle.autopower(
                minimum_frequency=1.5 / (xt.max() - xt.min()),
                maximum_frequency=0.25 * fsamp,
                samples_per_peak=10,
            )

            peak_freq = xf[np.argmax(yf)]
            peak_fap = lombscargle.false_alarm_probability(yf.max(), method="bootstrap")

            # output first iteration of LS periodogram
            if len(freqs) == 0:
                xf_out = xf.copy()
                yf_out = yf.copy()
                levels = lombscargle.false_alarm_level([0.1, 0.01, 0.001])

            if peak_fap < fap:
                yt -= lombscargle.model(xt, peak_freq) * len(xt)
                freqs.append(peak_freq)
                faps.append(peak_fap)

            else:
                loop = False

            if len(freqs) >= max_peaks:
                loop = False

        if return_levels:
            return xf_out, yf_out, freqs, faps, levels

        else:
            return xf_out, yf_out, freqs, faps


    def select_best_model(self, traces, dofs, verbose=True):
        """
        Arguments
            traces : dict
              dictionary of PyMC3 MultiTraces, model output from self.sample()
            dofs : dict
              degrees-of-freedom corresponding to each model trace

        Returns
            best_omc_model (str)
        """
        q = self.quality
        npts = np.sum(self.quality)

        # AIC and BIC for each model
        aic = {}
        bic = {}

        for k in traces.keys():
            trend = np.nanmedian(traces[k]['trend'], 0)
            lnlike = -np.sum((self.yobs[q] - trend)**2 / self.yerr[q]**2)

            aic[k] = 2 * dofs[k] - 2 * lnlike
            bic[k] = dofs[k] * np.log(npts) - 2 * lnlike

        preferred_by_aic = min(aic, key=aic.get)
        preferred_by_bic = min(bic, key=bic.get)

        if verbose:
            print(f"AIC : {preferred_by_aic}, BIC : {preferred_by_bic}")

        # Case 1: AIC and BIC match recommendation
        if preferred_by_aic == preferred_by_bic:
            best_omc_model = preferred_by_aic

        # Case 2: Select parametric models over Matern-3/2 GP
        elif np.any(np.array([preferred_by_aic, preferred_by_bic]) == 'matern32'):
            if preferred_by_aic == 'matern32':
                best_omc_model = preferred_by_bic
            elif preferred_by_bic == 'matern32':
                best_omc_model = preferred_by_aic
        
        # Case 3: Select model with more degrees of freedom
        else:
            if dofs[preferred_by_aic] >= dofs[preferred_by_bic]:
                best_omc_model = preferred_by_aic
            elif dofs[preferred_by_aic] < dofs[preferred_by_bic]:
                best_omc_model = preferred_by_bic

        return best_omc_model


    def calculate_outlier_probability(self, ymod):
        """
        Arguments:
            ymod (ndarray) : regularized model for self.yobs
        """
        # normalize and center residuals
        res = self.yobs - ymod
        res = res / np.std(res)
        res -= np.mean(res)

        # fit gaussian mixture
        with pm.Model() as model:
            w = pm.Dirichlet("w", np.array([1.0, 1.0]))
            mu = pm.Normal("mu", mu=0.0, sd=1.0, shape=1)
            tau_latent = pm.Gamma("tau_latent", 1.0, 1.0, shape=2)
            tau = pm.Deterministic("tau", T.sort(tau_latent))

            obs = pm.NormalMixture("obs", w, mu=mu * T.ones(2), tau=tau, observed=res)

        with model:
            trace = self.sample(model)

        # calculate foreground/background probability
        loc = np.nanmedian(trace["mu"], axis=0)
        scales = np.nanmedian(1 / np.sqrt(trace["tau"]), axis=0)

        z_fg = stats.norm(loc=loc, scale=scales.min()).pdf(res)
        z_bg = stats.norm(loc=loc, scale=scales.max()).pdf(res)

        fg_prob = z_fg / (z_fg + z_bg)

        # use KMeans clustering to assign each point to foreground/background
        km = KMeans(n_clusters=2)
        group = km.fit_predict(fg_prob.reshape(-1, 1))
        centroids = np.array([np.mean(fg_prob[group == 0]), np.mean(fg_prob[group == 1])])
        out = group == np.argmin(centroids)

        # reduce fg_prob threshold until less than 30% of points are flagged
        if np.sum(out) / len(out) > 0.3:
            out = fg_prob < np.percentile(fg_prob, 0.3)

        # return quality vector
        return fg_prob, out



__all__ = [
    "load_holczer_ttvs",
    "clean_holczer_ttvs",
    "match_holczer_ttv",
    "matern32_model",
    "poly_model",
    "sin_model",
    "mix_model",
    "flag_outliers",
]

import aesara_theano_fallback.tensor as T
from astropy.stats import mad_std
from celerite2.theano import GaussianProcess
from celerite2.theano import terms as GPterms
import pymc3 as pm
import pymc3_ext as pmx
import numpy as np
import numpy.polynomial.polynomial as poly
from scipy import stats
from scipy.ndimage import median_filter
from sklearn.cluster import KMeans


from .constants import *
from .utils import boxcar_smooth


def load_holczer_ttvs(file, npl, koi_id):
    """
    Docstring
    """
    data = np.loadtxt(file, usecols=[0, 1, 2, 3, 4])

    output = {}
    output["inds"] = [None] * npl
    output["tts"] = [None] * npl
    output["err"] = [None] * npl

    for n in range(npl):
        koi = int(koi_id[1:]) + 0.01 * (1 + n)
        use = np.isclose(data[:, 0], koi, rtol=1e-10, atol=1e-10)

        # Holczer uses BJD - 2454900; BJKD = BJD - 2454833
        if np.sum(use) > 0:
            output["inds"][n] = np.array(data[use, 1], dtype="int")
            output["tts"][n] = data[use, 2] + data[use, 3] / 24 / 60 + 67
            output["err"][n] = data[use, 4] / 24 / 60

    return output


def clean_holczer_ttvs(data, time_start, time_end, verbose=False):
    """
    Docstring
    """
    npl = len(data["tts"])

    data["period"] = np.nan * np.ones(npl)
    data["epoch"] = np.nan * np.ones(npl)
    data["full_inds"] = [None] * npl
    data["full_tts"] = [None] * npl
    data["out"] = [None] * npl

    for n in range(npl):
        if data["tts"][n] is not None:
            # fit a linear ephemeris
            pfit = poly.polyfit(data["inds"][n], data["tts"][n], 1)
            ephem = poly.polyval(data["inds"][n], pfit)

            # put fitted epoch in range (TIME_START, TIME_START + PERIOD)
            epoch, period = pfit

            if epoch < time_start:
                adj = 1 + (time_start - epoch) // period
                epoch += adj * period

            if epoch > (time_start + period):
                adj = (epoch - time_start) // period
                epoch -= adj * period

            data["period"][n] = np.copy(period)
            data["epoch"][n] = np.copy(epoch)

            full_ephem = np.arange(epoch, time_end, period)
            full_inds = np.array(np.round((full_ephem - epoch) / period), dtype="int")

            # calculate OMC and flag outliers
            xtime = np.copy(data["tts"][n])
            yomc = np.copy(data["tts"][n] - ephem)
            yerr = np.copy(data["err"][n])

            if len(yomc) > 16:
                ymed = boxcar_smooth(
                    median_filter(yomc, size=5, mode="mirror"), winsize=5
                )
            else:
                ymed = np.median(yomc)

            if len(yomc) > 4:
                out = np.abs(yomc - ymed) / mad_std(yomc - ymed) > 3.0
            else:
                out = np.zeros(len(yomc), dtype="bool")

            data["out"][n] = np.copy(out)

            # estimate TTV signal with a regularized Matern-3/2 GP
            if np.sum(~out) > 4:
                model = matern32_model(
                    xtime[~out], yomc[~out], yerr[~out], xt_predict=full_ephem
                )
            else:
                model = poly_model(
                    xtime[~out], yomc[~out], yerr[~out], 1, xt_predict=full_ephem
                )

            with model:
                map_soln = pmx.optimize(verbose=verbose)

            full_tts = full_ephem + map_soln["pred"]

            # track model prediction
            data["full_tts"][n] = np.copy(full_tts)
            data["full_inds"][n] = np.copy(full_inds)

    return data


def match_holczer_ttvs(planets, data):
    durs = np.zeros(len(planets))

    for n, p in enumerate(planets):
        durs[n] = p.duration

    for n, p in enumerate(planets):
        match = np.isclose(data["period"], p.period, rtol=0.1, atol=durs.max())

        if not ((np.sum(match) == 0) or (np.sum(match) == 1)):
            raise ValueError(
                "Something has gone wrong matching periods between DR25 and Holczer+2016"
            )

        if np.sum(match) == 1:
            loc = np.squeeze(np.where(match))

            hinds = data["inds"][loc]
            htts = data["tts"][loc]

            p.epoch = np.copy(data["epoch"][loc])
            p.period = np.copy(data["period"][loc])
            p.tts = np.copy(data["full_tts"][loc])
            p.index = np.copy(data["full_inds"][loc])

    return planets


def matern32_model(xtime, yomc, yerr, ymax=None, xt_predict=None):
    """
    Build a PyMC3 model to fit TTV observed-minus-calculated data
    Fits data with a regularized Matern-3/2 GP kernel

    Units (time) should be the same on all inputs

    Parameters
    ----------
    xtime : ndarray
        time values (e.g. linear ephemeris)
    yomc : ndarray
        observed-minus-caculated TTVs
    yerr : ndarray
        corresponding uncertainties on yomc
    ymax : float (optional)
        maximum allowed amplitude on GP scale parameter
    xt_predict : ndarray (optional)
        time values to predict OMC model; if not provided xtime will be used

    Returns
    -------
    model : pm.Model()
    """
    with pm.Model() as model:
        # delta between each transit time
        dx = np.mean(np.diff(xtime))

        # times where trend will be predicted
        if xt_predict is None:
            xt_predict = xtime

        # build the kernel
        if ymax is None:
            log_sigma = pm.Normal("log_sigma", mu=np.log(np.mean(yerr)), sd=5)
        else:
            log_sigma = pm.Bound(pm.Normal, upper=np.log(ymax))(
                "log_sigma", mu=np.log(ymax), sd=5
            )

        rho = pm.Uniform("rho", lower=2 * dx, upper=xtime.max() - xtime.min())
        kernel = GPterms.Matern32Term(sigma=T.exp(log_sigma), rho=rho)

        # mean
        mean = pm.Normal("mean", mu=np.mean(yomc), sd=np.std(yomc))

        # define the GP
        gp = GaussianProcess(kernel, t=xtime, yerr=yerr, mean=mean)

        # factorize the covariance matrix
        gp.compute(xtime, yerr=yerr)

        # track GP prediction, covariance, and degrees of freedom
        trend, cov = gp.predict(yomc, xtime, return_cov=True)

        trend = pm.Deterministic("trend", trend)
        cov = pm.Deterministic("cov", cov)
        pred = pm.Deterministic("pred", gp.predict(yomc, xt_predict))

        dof = pm.Deterministic("dof", pm.math.trace(cov / yerr**2))

        # add marginal likelihood to model
        lnlike = gp.log_likelihood(yomc)
        pm.Potential("lnlike", lnlike)

    return model


def poly_model(xtime, yomc, yerr, polyorder, xt_predict=None):
    """
    Build a PyMC3 model to fit TTV observed-minus-calculated data
    Fits data with a polynomial (up to cubic)

    Parameters
    ----------
    xtime : ndarray
        time values (e.g. linear ephemeris)
    yomc : ndarray
        observed-minus-caculated TTVs
    yerr : ndarray
        corresponding uncertainties on yomc
    polyorder : int
        polynomial order
    xt_predict : ndarray
        time values to predict OMC model; if not provided xtime will be used

    Returns
    -------
    model : pm.Model()
    """
    with pm.Model() as model:
        # times where trend will be predicted
        if xt_predict is None:
            xt_predict = xtime

        C0 = pm.Normal("C0", mu=0, sd=10)
        C1 = 0.0
        C2 = 0.0
        C3 = 0.0

        if polyorder >= 1:
            C1 = pm.Normal("C1", mu=0, sd=10)
        if polyorder >= 2:
            C2 = pm.Normal("C2", mu=0, sd=10)
        if polyorder >= 3:
            C3 = pm.Normal("C3", mu=0, sd=10)
        if polyorder >= 4:
            raise ValueError("only configured for 3rd order polynomials")

        def poly_fxn(c0, c1, c2, c3, xt):
            return c0 + c1 * xt + c2 * xt**2 + c3 * xt**3

        # mean
        trend = pm.Deterministic("trend", poly_fxn(C0, C1, C2, C3, xtime))

        # likelihood
        pm.Normal("obs", mu=trend, sd=yerr, observed=yomc)

        # track predicted trend
        pred = pm.Deterministic("pred", poly_fxn(C0, C1, C2, C3, xt_predict))

    return model


def sin_model(xtime, yomc, yerr, period, xt_predict=None):
    """
    Build a PyMC3 model to fit TTV observed-minus-calculated data
    Fits data with a single-frequency sinusoid

    Parameters
    ----------
    xtime : ndarray
        time values (e.g. linear ephemeris)
    yomc : ndarray
        observed-minus-caculated TTVs
    yerr : ndarray
        corresponding uncertainties on yomc
    period : float
        pre-estimated sinusoid period; the model places tight priors on this period
    xt_predict : ndarray
        time values to predict OMC model; if not provided xtime will be used

    Returns
    -------
    model : pm.Model()
    """
    with pm.Model() as model:
        # times where trend will be predicted
        if xt_predict is None:
            xt_predict = xtime

        # periodic component
        df = 1 / (xtime.max() - xtime.min())
        f = pm.Normal("f", mu=1 / period, sd=df)
        Ah = pm.Normal("Ah", mu=0, sd=5 * np.std(yomc))
        Bk = pm.Normal("Bk", mu=0, sd=5 * np.std(yomc))

        def sin_fxn(A, B, f, xt):
            return A * T.sin(2 * pi * f * xt) + B * T.cos(2 * pi * f * xt)

        # mean
        trend = pm.Deterministic("trend", sin_fxn(Ah, Bk, f, xtime))

        # likelihood
        pm.Normal("obs", mu=trend, sd=yerr, observed=yomc)

        # track predicted trend
        pred = pm.Deterministic("pred", sin_fxn(Ah, Bk, f, xt_predict))

    return model


def mix_model(x):
    """
    Build a 1D PyMC3 mixture model
    The model is composed of two normal distributions with the same mean but different variances

    Parameters
    ----------
    x : ndarray
        vector of data values

    Returns
    -------
    model : pm.Model()
    """
    xnorm = x / np.std(x)
    xnorm -= np.mean(xnorm)

    with pm.Model() as model:
        # mixture parameters
        w = pm.Dirichlet("w", np.array([1.0, 1.0]))
        mu = pm.Normal("mu", mu=0.0, sd=1.0, shape=1)
        tau = pm.Gamma("tau", 1.0, 1.0, shape=2)

        # here's the potential
        obs = pm.NormalMixture("obs", w, mu=mu * T.ones(2), tau=tau, observed=xnorm)

    return model


def flag_outliers(res, loc, scales):
    """
    Flag outliers in a residuals vector using a mixture model
    Applies unsupervised K-means clustering to assign points to either a narrow foreground or wide background distribution

    Parameters
    ----------
    res : ndarray (N)
        vector of residuals
    loc : float
        normal mean inferred from mix_model()
    scales : tuple
        normal standard deviations inferred from mix_model()

    Returns
    -------
    fg_prob : ndarray (N)
        probability the each item in res belongs to the foreground distribution
    out : bool (N)
        binary classification of each item in res into foreground/background distribution
    """
    resnorm = res / np.std(res)
    resnorm -= np.mean(resnorm)

    order = np.argsort(scales)
    scales = scales[order]

    # calculate foreground/background probability from results of mixture model
    z_fg = stats.norm(loc=loc, scale=scales[0]).pdf(resnorm)
    z_bg = stats.norm(loc=loc, scale=scales[1]).pdf(resnorm)

    fg_prob = z_fg / (z_fg + z_bg)

    # use KMeans clustering to assign each point to the foreground or background
    km = KMeans(n_clusters=2)
    group = km.fit_predict(fg_prob.reshape(-1, 1))
    centroids = np.array([np.mean(fg_prob[group == 0]), np.mean(fg_prob[group == 1])])
    out = group == np.argmin(centroids)

    return fg_prob, out



__all__ = ['SimpleDetrender',
           'GaussianProcessDetrender',
           'AutoCorrelationDetrender'
           ]


import aesara_theano_fallback.tensor as T
from astropy.stats import sigma_clip, mad_std
from astropy.timeseries import LombScargle
from celerite2.theano import GaussianProcess
from celerite2.theano import terms as GPterms
import numpy as np
import pymc3 as pm
import pymc3_ext as pmx
from scipy.signal import medfilt as median_filter
from scipy.signal import savgol_filter
from alderaan.planet import Planet
from alderaan.litecurve import LiteCurve
from alderaan.modules.base import BaseAlg

class SimpleDetrender(BaseAlg):
    def __init__(self, litecurve, planets):
        super().__init__(litecurve, planets)


    def clip_outliers(self, 
                      kernel_size, 
                      sigma_upper, 
                      sigma_lower, 
                      mask=None,
                      trend=None
                      ):
        """
        Sigma-clip outliers using astropy.stats.sigma_clip() and a median filtered trend
        Applies an additional iteration wrapper to allow for masked cadences

        Arguments:
            kernel_size : int
                size of the smoothing filter kernel
            sigma_upper : float
                upper sigmga clipping threshold
            sigma_lower : float
                lower sigma clipping threshold
            mask : array-like, bool (optional)
                do not reject cadences within masked regions; useful for protecting transits
            trend : array-like, float (optional)
                precomputed model for the trend, e.g. a Keplerian transit lightcurve
        """
        lc = self.litecurve

        if mask is None:
            mask = np.zeros(len(lc.time), dtype=bool)
        
        if trend is not None:
            raise ValueError("Not yet configured for trend kwarg")

        loop = True
        count = 0
        while loop:
            # first pass: try median filter
            if loop:
                smoothed = median_filter(lc.flux, kernel_size=kernel_size)

                bad = sigma_clip(lc.flux - smoothed,
                                 sigma_upper=sigma_upper,
                                 sigma_lower=sigma_lower,
                                 stdfunc=mad_std
                                ).mask
                bad = bad * ~mask

            # second pass: try savgol filter
            # if over 1% of points were flagged with median filter
            if np.sum(bad)/len(bad) > 0.01:
                smoothed = savgol_filter(lc.flux,
                                         window_length=kernel_size,
                                         polyorder=2
                                        )
                bad = sigma_clip(lc.flux - smoothed,
                                 sigma_upper=sigma_upper,
                                 sigma_lower=sigma_lower,
                                 stdfunc=mad_std
                                ).mask
                bad = bad * ~mask

            # third pass: skip outlier rejection
            if np.sum(bad) / len(bad) > 0.01:
                bad = np.zeros_like(mask)

            # set attributes
            for k in lc.__dict__.keys():
                if type(lc.__dict__[k]) is np.ndarray:
                    lc.__setattr__(k, lc.__dict__[k][~bad])

            # update mask and iterate
            mask = mask[~bad]

            if np.sum(bad) == 0:
                loop = False
            else:
                count += 1

            if count >= 3:
                loop = False

        self.litecurve = lc


    def _identify_breaks(self, gap_tolerance, jump_tolerance, mask=None):
        """
        Identify gaps (breaks in times) and jumps (sudden flux changes) in a litecurve
            * GAPS are consecutive missing cadences in the TIME domain
            * JUMPS are large cadence-to-cadence FLUX variations

        Arguments
            gap_tolerance : int
                number of cadences to be considered a (time) gap
            jump_tolerance : float
                sigma threshold for identifying (flux) jumps
            mask : array-like, bool (optional)
                1 near transit, 0 far from transit
        """
        # identify GAPS in time
        gaps = np.pad(np.diff(self.litecurve.cadno), (1,0), 'constant', constant_values=(1,0))
        gap_locs = np.pad(
            np.where(gaps > gap_tolerance)[0], (1,1), 'constant', constant_values=(0, len(gaps)+1)
        )

        # identify JUMPS in flux
        jumps = np.pad(np.diff(self.litecurve.flux), (1,0), 'constant', constant_values=(0,0))
        jump_size = np.abs(jumps - np.median(jumps)) / mad_std(jumps)
        jump_locs = np.where(~mask.astype(bool) & (jump_size > jump_tolerance))[0]

        # combine breaks
        break_locs = np.sort(np.unique(np.hstack([gap_locs, jump_locs])))
        
        # fix edge behavior
        bad = np.hstack([False, np.diff(break_locs) < gap_tolerance])
        if bad[-1]:
            bad[-2:] = (True, False)
        break_locs = break_locs[~bad]
        break_locs[-1] -= 2

        assert break_locs[0] == 0, "unexpected edge behavior"
        assert break_locs[-1] == len(self.litecurve.cadno) - 1, "unexpected edge behavior"
        
        return break_locs


class GaussianProcessDetrender(SimpleDetrender):
    def __init__(self, litecurve, planets):
        super().__init__(litecurve, planets)


    def estimate_oscillation_period(self, min_period=None):
        """
        Docstring
        """
        lc = self.litecurve
        lombscargle = LombScargle(lc.time, lc.flux)

        if min_period is None:
            min_period = 3 * np.max(self.durs)

        min_freq = 1 / (lc.time.max() - lc.time.min())
        max_freq = 1 / min_period

        xf, yf = lombscargle.autopower(
            minimum_frequency=min_freq, maximum_frequency=max_freq
        )

        peak_freq = xf[np.argmax(yf)]
        peak_per = np.max([1.0 / peak_freq, 1.001 * min_period])

        return peak_per
    

    def detrend(self,
                gp_term,
                nominal_gp_period,
                minimum_gp_period,
                transit_mask=None,
                gap_tolerance=None,
                jump_tolerance=None,
                correct_ramp=True,
                return_trend=False, 
                progressbar=False
               ):
        """
        Arguments
            gp_term : str "RotationTerm" or "SHOTerm"
            nominal_gp_period : float
            minimum_gp_period : float
            transit_mask : ndarray
            gap_tolerance : int
            jump_tolerance : float
            correct_ramp : bool
            return_trend : bool
            progressbar : bool
        """
        # shorthand
        lc = self.litecurve

        # sanitize inputs
        if transit_mask is None:
            transit_mask = np.zeros(len(lc.time), dtype=bool)
        if gap_tolerance is None:
            gap_tolerance = len(lc.time)
        if jump_tolerance is None:
            jump_tolerance = 1000.

        # identify breaks (gaps in time, jumps in flux)
        # these will be used to identify contiguous "segments" of data
        # the model will fit all n_seg segments simultaneously
        # mean function (flux baseline + exponential trend) is independent for each segment
        # gp parameters are shared between all segments
        
        break_locs = self._identify_breaks(gap_tolerance, jump_tolerance, mask=transit_mask)
        
        n_seg = len(break_locs) - 1
        seg_id = np.zeros(len(lc.time), dtype=int)

        for i in range(n_seg):
            seg_id[break_locs[i]:break_locs[i+1]] = i

        approx_mean = [np.mean(lc.flux[seg_id == i]) for i in range(n_seg)]

        # convenience function
        def _mean_fxn(time, seg_id, flux0, ramp_amp, log_tau):
            nseg = len(np.unique(seg_id))
            mean = T.zeros(len(time))

            for i in range(nseg):
                t_min = time[seg_id == i].min()
                mean += (
                    flux0[i]
                    * (1 + ramp_amp[i] * T.exp(-(time - t_min) / T.exp(log_tau[i])))
                    * (seg_id == i)
                )

            return mean
        
        # pymc model
        with pm.Model() as model:
            # set up the kernel
            log_sigma = pm.Normal("log_sigma", mu=np.log(mad_std(lc.flux)), sd=5.0)
            sigma = pm.Deterministic("sigma", T.exp(log_sigma))

            log_P_off = pm.Normal("log_P", mu=np.log(nominal_gp_period - minimum_gp_period), sd=2.0)
            P = pm.Deterministic("P", minimum_gp_period + T.exp(log_P_off))

            log_Q0 = pm.Normal("log_Q0", mu=0.0, sd=5.0, testval=np.log(0.5))
            Q0 = pm.Deterministic("Q0", T.exp(log_Q0))

            if gp_term == 'RotationTerm':
                log_dQ = pm.Normal("log_dQ", mu=0.0, sd=5.0, testval=np.log(1e-3))
                mix = pm.Uniform("mix", lower=0, upper=1, testval=0.1)
                kernel = GPterms.RotationTerm(sigma=sigma, period=P, Q0=Q0, dQ=T.exp(log_dQ), f=mix)

            elif gp_term == 'SHOTerm':
                kernel = GPterms.SHOTerm(sigma=sigma, w0=2 * pi / P, Q=0.5 + Q0)
                
            else:
                raise ValueError(f"gp_term {gp_term} not recognized")
            
            # set up mean function and variance
            flux0 = pm.Normal("flux0", mu=approx_mean, sd=np.ones(n_seg), shape=n_seg)
            
            if correct_ramp:
                ramp_amp = pm.Normal("ramp_amp", mu=0, sd=np.std(lc.flux), shape=n_seg)
                log_tau = pm.Normal("log_tau", mu=0, sd=5, shape=n_seg)
            else:
                ramp_amp = T.zeros(n_seg)
                log_tau = T.zeros(n_seg)

            # set up the gp
            oot = ~transit_mask
            mean_oot = pm.Deterministic("mean_oot", _mean_fxn(lc.time[oot], seg_id[oot], flux0, ramp_amp, log_tau))
            mean_all = pm.Deterministic("mean_all", _mean_fxn(lc.time, seg_id, flux0, ramp_amp, log_tau))
            log_var = pm.Normal("log_var", mu=np.var(lc.flux - median_filter(lc.flux, 13)), sd=5.0)

            gp = GaussianProcess(kernel, mean=mean_oot)
            gp.compute(lc.time[oot], diag=T.exp(log_var)*T.ones(len(lc.time[oot])))
            
            # likelihood
            pm.Potential("lnlike", gp.log_likelihood(lc.flux[oot]))

            # predicted trend; if len(y)!=len(t) set include_mean=False
            pred = gp.predict(lc.flux[oot], t=lc.time, include_mean=False)
            pm.Deterministic("pred", pred + mean_all)

        # optimize hyperparameters
        with model:
            map_soln = pmx.optimize(start=model.test_point, vars=[flux0], progress=progressbar)
            map_soln = pmx.optimize(start=map_soln, vars=[flux0, log_var], progress=progressbar)

            for i in range(1 + correct_ramp):
                if gp_term == 'RotationTerm':
                    map_soln = pmx.optimize(
                        start=map_soln,
                        vars=[log_var, flux0, sigma, P, log_Q0, log_dQ, mix],
                        progress=progressbar,
                    )
                if gp_term == 'SHOTerm':
                    map_soln = pmx.optimize(
                        start=map_soln,
                        vars=[log_var, flux0, sigma, P, log_Q0],
                        progress=progressbar,
                    )
                if correct_ramp:
                    map_soln = pmx.optimize(
                        start=map_soln,
                        vars=[log_var, flux0, ramp_amp, log_tau],
                        progress=progressbar,
                    )

        self.litecurve.flux /= map_soln['pred']
        self.litecurve.error /= map_soln['pred']

        if return_trend:
            return self.litecurve, map_soln['pred']
        return self.litecurve


class AutoCorrelationDetrender(SimpleDetrender):
    def __init__(self, litecurve, planets):
        super().__init__(litecurve, planets)



import os
import sys

from astropy.units import UnitsWarning
from pathlib import Path
import warnings
from alderaan.litecurve import LiteCurve


warnings.simplefilter('always', UserWarning)
warnings.filterwarnings(
    action='ignore', category=UnitsWarning, module='astropy'
)

base_path = Path(__file__).resolve().parents[1]
data_dir = os.path.join(base_path, 'alderaan/examples/data/MAST_downloads/')
kic_id = 8644288  # KOI-137 (Kepler-18)

litecurve = LiteCurve().from_kplr_pdcsap(data_dir, kic_id, 'long cadence')

print("\npassing")


__all__ = [
    "parse_catalog",
    "read_mast_files",
    "cleanup_lkc",
    "LightKurve_to_LiteCurve",
    "transit_parameters_to_dataframe",
    "load_detrended_lightcurve",
    "dynesty_results_to_fits",
]


import os
from copy import deepcopy

from astropy.io import fits
import lightkurve as lk
import numpy as np
import pandas as pd

from .LiteCurve import LiteCurve


def parse_catalog(file, mission, target):
    """
    Parse catalog data for a single star given an input csv file

    Parameters
    ----------
        file : str
            path to input csv file with star and planet properties
        mission : str
            can be 'Kepler' or 'Kepler-Validation'
        target : str
            target id, e.g. K00137

    Returns
    -------
        catalog : pd.DataFrame
            dataframe with planet and star properties, sorted by ascending period

    """
    catalog = pd.read_csv(file, index_col=0)
    catalog = catalog.loc[catalog.koi_id == target]

    # check for consistency in multi-planet systems
    if not all(kic_id == catalog.kic_id.to_numpy()[0] for kic_id in catalog.kic_id):
        raise ValueError("There are inconsistencies with KIC in the csv input file")

    if not all(npl == catalog.npl.to_numpy()[0] for npl in catalog.npl):
        raise ValueError("There are inconsistencies with NPL in the csv input file")

    if not all(
        ld_u1 == catalog.limbdark_1.to_numpy()[0] for ld_u1 in catalog.limbdark_1
    ):
        raise ValueError("There are inconsistencies with LD_U1 in the csv input file")

    if not all(
        ld_u2 == catalog.limbdark_2.to_numpy()[0] for ld_u2 in catalog.limbdark_2
    ):
        raise ValueError("There are inconsistencies with LD_U2 in the csv input file")

    # check for NaN valued transit parameters
    if np.any(
        np.isnan(np.array(catalog["period epoch depth duration impact".split()]))
    ):
        raise ValueError("NaN values found in input catalog")

    # sort by ascending period
    catalog = catalog.sort_values(by="period").reset_index(drop=True)

    return catalog


def read_mast_files(mast_files, kic_id, obsmode, exclude=None):
    """
    Docstring
    """
    if exclude is None:
        exclude = []

    rawdata_list = []
    for i, mf in enumerate(mast_files):
        with fits.open(mf) as hdu_list:
            if hdu_list[0].header["OBSMODE"] == obsmode and ~np.isin(
                hdu_list[0].header["QUARTER"], exclude
            ):
                rawdata_list.append(lk.read(mf))

    lkcol = cleanup_lkc(lk.LightCurveCollection(rawdata_list), kic_id)

    return lkcol


def cleanup_lkc(lk_collection, kic):
    """
    Join each quarter in a lk.LightCurveCollection into a single lk.LightCurve
    Performs only the minimal detrending step remove_nans()

    Parameters
    ----------
        lk_collection : lk.LightCurveCollection
            lk.LightCurveCollection() with (possibly) multiple entries per quarter
        kic : int
            Kepler Input Catalogue (KIC) number for target

    Returns
    -------
        lkc : lk.LightCurveCollection
            lk.LightCurveCollection() with only one entry per quarter
    """
    lk_col = deepcopy(lk_collection)

    quarters = []
    for lkc in lk_col:
        quarters.append(lkc.quarter)

    data_out = []
    for q in np.unique(quarters):
        lkc_list = []
        cadno = []

        for lkc in lk_col:
            if (lkc.quarter == q) * (lkc.targetid == kic):
                lkc_list.append(lkc)
                cadno.append(lkc.cadenceno.min())

        order = np.argsort(cadno)
        lkc_list = [lkc_list[j] for j in order]

        # the operation "stitch" converts a LightCurveCollection to a single LightCurve
        lkc = lk.LightCurveCollection(lkc_list).stitch().remove_nans()

        data_out.append(lkc)

    return lk.LightCurveCollection(data_out)


def LightKurve_to_LiteCurve(lklc):
    return LiteCurve(
        time=np.array(lklc.time.value, dtype="float"),
        flux=np.array(lklc.flux.value, dtype="float"),
        error=np.array(lklc.flux_err.value, dtype="float"),
        cadno=np.array(lklc.cadenceno.value, dtype="int"),
        quarter=lklc.quarter * np.ones(len(lklc.time), dtype="int"),
        season=(lklc.quarter % 4) * np.ones(len(lklc.time), dtype="int"),
        quality=lklc.quality.value,
    )


def transit_parameters_to_dataframe(koi_id, kic_id, planets, limbdark):
    npl = len(planets)

    data = {}
    keys = "koi_id kic_id npl period epoch depth duration impact limbdark_1 limbdark_2".split()

    for k in keys:
        data[k] = [None] * npl

    for n, p in enumerate(planets):
        data["koi_id"][n] = koi_id
        data["kic_id"][n] = kic_id
        data["npl"][n] = npl
        data["period"][n] = p.period
        data["epoch"][n] = p.epoch
        data["depth"][n] = p.depth * 1e6
        data["duration"][n] = p.duration * 24
        data["impact"][n] = p.impact
        data["limbdark_1"][n] = limbdark[0]
        data["limbdark_2"][n] = limbdark[1]

    return pd.DataFrame(data)


def load_detrended_lightcurve(filename):
    """
    Load a fits file previously generated by LiteCurve.to_fits()

    Parameters
    ----------
        filename : string

    Returns
    -------
        litecurve : LiteCurve() object

    """
    litecurve = LiteCurve()

    with fits.open(filename) as hdulist:
        litecurve.time = np.array(hdulist["TIME"].data, dtype="float64")
        litecurve.flux = np.array(hdulist["FLUX"].data, dtype="float64")
        litecurve.error = np.array(hdulist["ERROR"].data, dtype="float64")
        litecurve.cadno = np.array(hdulist["CADNO"].data, dtype="int")
        litecurve.quarter = np.array(hdulist["QUARTER"].data, dtype="int")

    return litecurve


def dynesty_results_to_fits(results, project_dir, mission, target, run_id):
    """
    results : dynesty.DynamicNestedSampling.results
    target : (str) name of target, e.g. 'K00137'
    """
    npl = (results.samples.shape[1] - 2) // 5

    # package nested samples
    samples_keys = []

    for n in range(npl):
        samples_keys += "C0_{0} C1_{0} ROR_{0} IMPACT_{0} DUR14_{0}".format(n).split()

    samples_keys += ["LD_Q1", "LD_Q2"]
    samples_keys += ["LN_WT", "LN_LIKE", "LN_Z"]

    samples_data = np.vstack(
        [results.samples.T, results.logwt, results.logl, results.logz]
    ).T
    samples_df = pd.DataFrame(samples_data, columns=samples_keys)

    # primary HDU
    primary_hdu = fits.PrimaryHDU()
    primary_hdu.header["MISSION"] = mission
    primary_hdu.header["TARGET"] = target
    primary_hdu.header["RUN_ID"] = run_id
    primary_hdu.header["NPL"] = npl

    # samples HDU
    samples_hdu = fits.BinTableHDU(
        data=samples_df.to_records(index=False), name="SAMPLES"
    )

    samples_hdu.header["NITER"] = results.niter
    samples_hdu.header["NBATCH"] = len(results.batch_nlive)
    for i, nlive in enumerate(results.batch_nlive):
        samples_hdu.header[f"NLIVE{i}"] = nlive
    samples_hdu.header["EFF"] = results.eff

    # build HDU List
    hduL = fits.HDUList([primary_hdu, samples_hdu])

    # add transit times to HDU List
    for n in range(npl):
        ttimes_file = os.path.join(
            project_dir,
            f"Results/{run_id}/{target}/{target}_{str(n).zfill(2)}_quick.ttvs",
        )
        ttimes_keys = "INDEX TTIME MODEL OUT_PROB OUT_FLAG".split()
        ttimes_data = np.loadtxt(ttimes_file)

        ttimes_df = pd.DataFrame(ttimes_data, columns=ttimes_keys)
        ttimes_df.INDEX = ttimes_df.INDEX.astype("int")
        ttimes_df.OUT_FLAG = ttimes_df.OUT_FLAG.astype("int")

        ttimes_hdu = fits.BinTableHDU(
            data=ttimes_df.to_records(index=False),
            name=f"TTIMES_{str(n).zfill(2)}",
        )

        hduL.append(ttimes_hdu)

    return hduL



__all__ = ['uniform_ppf',
           'loguniform_ppf',
           'norm_ppf',
           ]


import numpy as np
from scipy.special import erfinv


def uniform_ppf(u, a, b):
    """
    Transform from U(0,1) --> Uniform(a,b)

    Args:
        u (array-like) : samples from standard uniform distribution U(0,1)
        a (float) : lower bound on transformed distribution U(a,b)
        b (float) : upper bound on transformed distribution U(a,b)

    Returns:
        array-like : transformed samples from uniform distribution
    """
    return u * (b - a) + a


def loguniform_ppf(u, a, b):
    """
    Transform from U(0,1) --> LogUniform(a,b)

    Args:
        u (array-like) : samples from standard uniform distribution U(0,1)
        a (float) : lower bound on transformed distribution lnU(a,b)
        b (float) : upper bound on transformed distribution lnU(a,b)

    Returns:
        x (array-like) : transformed samples from log-uniform distribution
    """
    return np.exp(u * np.log(b) + (1 - u) * np.log(a))


def norm_ppf(u, mu, sig, eps=1e-12):
    """
    Transform from U(0,1) --> Normal(mu,sig)

    Args:
        u (array-like) : samples from standard uniform distribution U(0,1)
        mu (float) : mean of transformed distribution N(mu,sig)
        sig (float) : standard deviation of transformed distribution N(mu,sig)
        eps (float) : provides numerical stability in the inverse error function at edges

    Returns
    -------
        array-like : transformed samples from Normal distribution
    """
    return mu + sig * np.sqrt(2) * erfinv((2 * u - 1) / (1 + eps))



import astropy.constants as apc
import numpy as np

pi = np.pi

RJRE = (apc.R_jup / apc.R_earth).value  # R_jup/R_earth
RSRE = (apc.R_sun / apc.R_earth).value  # R_sun/R_earth
RSRJ = RSRE / RJRE  # R_sun/R_jup

MJME = (apc.M_jup / apc.M_earth).value  # M_jup/M_earth
MSME = (apc.M_sun / apc.M_earth).value  # M_sun/M_earth
MSMJ = MSME / MJME  # M_sun/M_jup

RSAU = (apc.R_sun / apc.au).value  # solar radius [AU]
RHOSUN_GCM3 = (
    3 * apc.M_sun / (4 * pi * apc.R_sun**3)
).value / 1000  # solar density [g/cm^3]

kepler_lcit = 29.4243885 / 60 / 24  # Kepler long cadence integration time + readout time [days]
kepler_scit = 58.848777 / 3600 / 24 # Kepler short cadence integration time + readout time [days]


#!/usr/bin/env python
# coding: utf-8

# # Fit transit shape with nested sampling


import os
import sys
import json
import shutil
import warnings
import multiprocessing as multipro
from datetime import datetime
from timeit import default_timer as timer

import argparse
import batman
import dynesty
from dynesty import plotting as dyplot
import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy.polynomial.polynomial as poly
import numpy as np

from aesara_theano_fallback import aesara as theano
from celerite2 import terms as GPterms


# #### Flush buffer and silence extraneous warnings


# flush buffer to avoid mixed outputs from progressbar
sys.stdout.flush()

# turn off FutureWarnings
warnings.filterwarnings("ignore", category=FutureWarning)


# #### Initialize timer


print("")
print("+" * shutil.get_terminal_size().columns)
print("ALDERAAN Detrending and TTV Estimation")
print(f"Initialized {datetime.now().strftime('%d-%b-%Y at %H:%M:%S')}")
print("+" * shutil.get_terminal_size().columns)
print("")

# start program timer
global_start_time = timer()


# #### Parse inputs


try:
    parser = argparse.ArgumentParser(
        description="Inputs for ALDERAAN transit fiting pipeline"
    )
    parser.add_argument(
        "--mission",
        default=None,
        type=str,
        required=True,
        help="Mission name; can be 'Kepler' or 'Simulated'",
    )
    parser.add_argument(
        "--target",
        default=None,
        type=str,
        required=True,
        help="Target name; format should be K00000 or S00000",
    )
    parser.add_argument(
        "--run_id", default=None, type=str, required=True, help="run identifier"
    )
    parser.add_argument(
        "--project_dir",
        default=None,
        type=str,
        required=True,
        help="Project directory for accessing lightcurve data and saving outputs",
    )
    parser.add_argument(
        "--verbose",
        default=False,
        type=bool,
        required=False,
        help="'True' to enable verbose logging",
    )
    parser.add_argument(
        "--iplot",
        default=False,
        type=bool,
        required=False,
        help="'True' to enable interactive matplotlib backend; default 'agg'",
    )

    args = parser.parse_args()
    MISSION = args.mission
    TARGET = args.target
    RUN_ID = args.run_id
    PROJECT_DIR = args.project_dir
    VERBOSE = args.verbose
    IPLOT = args.iplot

except SystemExit:
    warnings.warn("No arguments were parsed from the command line")


print("")
print(f"   MISSION : {MISSION}")
print(f"   TARGET  : {TARGET}")
print(f"   RUN ID  : {RUN_ID}")
print("")
print(f"   Project directory : {PROJECT_DIR}")
print("")
print(f"   theano cache : {theano.config.compiledir}")
print("")


# #### Build directory structure


# directories in which to place pipeline outputs for this run
RESULTS_DIR = os.path.join(PROJECT_DIR, "Results", RUN_ID, TARGET)
FIGURE_DIR = os.path.join(PROJECT_DIR, "Figures", RUN_ID, TARGET)

os.makedirs(RESULTS_DIR, exist_ok=True)
os.makedirs(FIGURE_DIR, exist_ok=True)

sys.path.append(PROJECT_DIR)


# #### Import ALDERAAN routines


import alderaan.io as io
import alderaan.dynesty_helpers as dynhelp
from alderaan.astro import make_transit_mask, set_oversample_factor
from alderaan.constants import scit, lcit
from alderaan.Ephemeris import Ephemeris
from alderaan.plotting import plot_omc


# #### Set matplotlib backend


# check for interactive matplotlib backends
if not IPLOT:
    mpl.use("agg")

if np.any(np.array(["agg", "png", "svg", "pdf", "ps"]) == mpl.get_backend()):
    IPLOT = False

# flush buffer
sys.stdout.flush()

# MAIN SCRIPT BEGINS HERE
def main():
    # # ################
    # # ----- DATA I/O -----
    # # ################

    print("\nLoading data...\n")

    # ## Read in planet and stellar properties

    path = os.path.join(RESULTS_DIR, f"{TARGET}_transit_parameters.csv")
    catalog = io.parse_catalog(path, MISSION, TARGET)

    KOI_ID = catalog.koi_id.to_numpy()[0]
    KIC_ID = catalog.kic_id.to_numpy()[0]

    NPL = catalog.npl.to_numpy()[0]

    PERIODS = catalog.period.to_numpy()
    DEPTHS = catalog.depth.to_numpy() * 1e-6
    DURS = catalog.duration.to_numpy() / 24

    U1 = catalog.limbdark_1.to_numpy()[0]
    U2 = catalog.limbdark_2.to_numpy()[0]

    if np.any(np.diff(PERIODS) <= 0):
        raise ValueError("Planets should be ordered by ascending period")

    # ## Read in filtered lightcurves
    # #### These can be generated by running the script "analyze_autocorrelated_noise.py"

    if os.path.exists(os.path.join(RESULTS_DIR, f"{TARGET}_lc_filtered.fits")):
        lc = io.load_detrended_lightcurve(
            os.path.join(RESULTS_DIR, f"{TARGET}_lc_filtered.fits")
        )
        lc.season = lc.quarter % 4
    else:
        lc = None

    if os.path.exists(os.path.join(RESULTS_DIR, f"{TARGET}_sc_filtered.fits")):
        sc = io.load_detrended_lightcurve(
            os.path.join(RESULTS_DIR, f"{TARGET}_sc_filtered.fits")
        )
        sc.season = sc.quarter % 4
    else:
        sc = None

    # ## Read in quick transit times
    # #### These can be generated by running the script "detrend_and_estimate_ttvs.py"

    # transit times
    epochs = np.zeros(NPL)
    periods = np.zeros(NPL)
    linear_ephemeris = [None] * NPL

    transit_inds = []
    indep_transit_times = []
    regular_transit_times = []

    for n in range(NPL):
        fname_in = os.path.join(RESULTS_DIR, f"{TARGET}_{n:02d}_quick.ttvs")
        data_in = np.genfromtxt(fname_in)

        transit_inds.append(np.array(data_in[:, 0], dtype="int"))
        indep_transit_times.append(np.array(data_in[:, 1], dtype="float"))
        regular_transit_times.append(np.array(data_in[:, 2], dtype="float"))

        # do a quick fit to get a linear ephemeris
        pfit = poly.polyfit(transit_inds[n], regular_transit_times[n], 1)

        epochs[n] = pfit[1]
        periods[n] = pfit[0]
        linear_ephemeris[n] = poly.polyval(transit_inds[n], pfit)

    # calculate centered transit indexes
    centered_transit_inds = [None] * NPL
    for n in range(NPL):
        centered_transit_inds[n] = transit_inds[n] - transit_inds[n][-1] // 2

    # make OMC plot
    fig, ax = plot_omc(linear_ephemeris, indep_transit_times, regular_transit_times)
    fig.savefig(os.path.join(FIGURE_DIR, TARGET + f"_omc.png"), bbox_inches="tight")
    if IPLOT:
        plt.show()
    else:
        plt.close()

    # # ####################
    # # --- PRELIMINARIES ---
    # # ####################

    print("\nRunning preliminaries...\n")

    # ## Establish time baseline

    time_min = []
    time_max = []

    if sc is not None:
        time_min.append(sc.time.min())
        time_max.append(sc.time.max())

    if lc is not None:
        time_min.append(lc.time.min())
        time_max.append(lc.time.max())

    TIME_START = np.min(time_min)
    TIME_END = np.max(time_max)

    # put epochs in range (TIME_START, TIME_START + PERIOD)
    for n in range(NPL):
        if epochs[n] < TIME_START:
            adj = 1 + (TIME_START - epochs[n]) // periods[n]
            epochs[n] += adj * periods[n]

        if epochs[n] > (TIME_START + periods[n]):
            adj = (epochs[n] - TIME_START) // periods[n]
            epochs[n] -= adj * periods[n]

    # ## Make transit masks

    if sc is not None:
        sc_mask = np.zeros((NPL, len(sc.time)), dtype="bool")
        for n in range(NPL):
            sc_mask[n] = make_transit_mask(
                sc.time,
                regular_transit_times[n],
                masksize=np.max([1 / 24, 1.5 * DURS[n]]),
            )

    if lc is not None:
        lc_mask = np.zeros((NPL, len(lc.time)), dtype="bool")
        for n in range(NPL):
            lc_mask[n] = make_transit_mask(
                lc.time,
                regular_transit_times[n],
                masksize=np.max([1 / 24, 1.5 * DURS[n]]),
            )

    # ## Identify overlapping transits

    overlap = [None] * NPL

    for i in range(NPL):
        overlap[i] = np.zeros(len(linear_ephemeris[i]), dtype="bool")

        for j in range(NPL):
            if i != j:
                for tt in linear_ephemeris[j]:
                    overlap[i] += np.abs(linear_ephemeris[i] - tt) < (
                        DURS[i] + DURS[j] + lcit
                    )

    # ## Predict number of points expected in transit

    # number of photometric points expected in transit
    count_expect_lc = np.zeros(NPL, dtype="int")
    count_expect_sc = np.zeros(NPL, dtype="int")

    for n in range(NPL):
        count_expect_lc[n] = np.max([1, int(np.floor(DURS[n] / lcit))])
        count_expect_sc[n] = np.max([1, int(np.floor(DURS[n] / scit))])

    # ## Grab data near transits

    # go quarter-by-quarter
    all_time = [None] * 18
    all_flux = [None] * 18
    all_error = [None] * 18
    all_mask = [None] * 18
    all_dtype = ["none"] * 18

    for q in range(18):
        if sc is not None:
            if np.isin(q, sc.quarter):
                use = (sc_mask.sum(0) != 0) * (sc.quarter == q)

                if np.sum(use) > np.min(count_expect_sc):
                    all_time[q] = sc.time[use]
                    all_flux[q] = sc.flux[use]
                    all_error[q] = sc.error[use]
                    all_mask[q] = sc_mask[:, use]
                    all_dtype[q] = "short"

                else:
                    all_dtype[q] = "short_no_transits"

        if lc is not None:
            if np.isin(q, lc.quarter):
                use = (lc_mask.sum(0) != 0) * (lc.quarter == q)

                if np.sum(use) > np.min(count_expect_lc):
                    all_time[q] = lc.time[use]
                    all_flux[q] = lc.flux[use]
                    all_error[q] = lc.error[use]
                    all_mask[q] = lc_mask[:, use]
                    all_dtype[q] = "long"

                else:
                    all_dtype[q] = "long_no_transits"

    all_dtype = np.array(all_dtype)

    # track which quarters have coverage
    quarters = np.arange(18)[(all_dtype == "short") + (all_dtype == "long")]

    # ## Track which quarter each transit falls in

    # get list of threshold times between quarters
    qthresh = np.zeros(len(quarters) + 1)
    qthresh[0] = TIME_START - 0.5

    for j, q in enumerate(quarters):
        if lc is not None:
            if np.isin(q, np.unique(lc.quarter)):
                qthresh[j + 1] = lc.time[lc.quarter == q].max() + lcit

        if sc is not None:
            if np.isin(q, np.unique(sc.quarter)):
                qthresh[j + 1] = sc.time[sc.quarter == q].max() + scit

    qthresh[-1] = TIME_END + 0.5

    # track which quarter each transit falls in
    transit_quarter = [None] * NPL

    for n in range(NPL):
        tts = regular_transit_times[n]
        transit_quarter[n] = np.zeros(len(tts), dtype="int")

        for j, q in enumerate(quarters):
            transit_quarter[n][(tts >= qthresh[j]) * (tts < qthresh[j + 1])] = q

    # ## Track mean and variance of each quarter

    # track mean and variance of each quarter
    mean_by_quarter = np.ones(18) * np.nan
    var_by_quarter = np.ones(18) * np.nan

    for q in range(18):
        if sc is not None:
            if np.isin(q, sc.quarter):
                mean_by_quarter[q] = np.mean(sc.flux[sc.quarter == q])
                var_by_quarter[q] = np.var(sc.flux[sc.quarter == q])

        if lc is not None:
            if np.isin(q, lc.quarter):
                mean_by_quarter[q] = np.mean(lc.flux[lc.quarter == q])
                var_by_quarter[q] = np.var(lc.flux[lc.quarter == q])

    # ## Set oversampling factor and exposure times

    # calculate oversample factor
    oversample_lc = set_oversample_factor(
        PERIODS, DEPTHS, DURS, lc.flux, lc.error, lcit
    )
    oversample_sc = 1

    # broadcast oversample factor
    oversample = np.zeros(18, dtype="int")
    oversample[all_dtype == "short"] = oversample_sc
    oversample[all_dtype == "long"] = oversample_lc

    # broadcast exposure times
    texp = np.zeros(18)
    texp[all_dtype == "short"] = scit
    texp[all_dtype == "long"] = lcit

    # precompute exposure integration time offsets
    texp_offsets = [None] * 18
    for j, q in enumerate(quarters):
        if all_dtype[q] == "short":
            texp_offsets[q] = np.array([0.0])
        elif all_dtype[q] == "long":
            texp_offsets[q] = np.linspace(-texp[q] / 2.0, texp[q] / 2.0, oversample[q])

    # ## Set GP noise priors

    # Read in noise model GP priors from analyze_autocorrelated_noise.py
    gp_percs = []

    for z in range(4):
        try:
            fname_in = os.path.join(RESULTS_DIR, f"{TARGET}_noise_gp_priors_{z}.txt")

            with open(fname_in) as infile:
                gp_percs.append(json.load(infile))

        except Exception:
            gp_percs.append(None)

    # convert the percentile priors into Gaussians
    gp_priors = []

    for z in range(4):
        if gp_percs[z] is not None:
            gpz = {}

            for k in gp_percs[z].keys():
                if k != "percentiles":
                    perc = np.array(gp_percs[z]["percentiles"])

                    med = np.array(gp_percs[z][k])[perc == 50.0][0]
                    err1 = np.array(gp_percs[z][k])[perc == 84.135][0]
                    err2 = np.array(gp_percs[z][k])[perc == 15.865][0]

                    dev = np.sqrt((err1 - med) ** 2 / 2 + (err2 - med) ** 2 / 2)

                    gpz[k] = (med, dev)

            gp_priors.append(gpz)

        else:
            # these are dummy values that effectively create a zero-amplitude kernel
            gpz = {}
            gpz["logw0"] = [np.log(2 * np.pi / (7 * DURS.max()))]
            gpz["logSw4"] = [-100.0]
            gpz["logQ"] = [np.log(1 / np.sqrt(2))]

            gp_priors.append(gpz)

    # calculate a few convenience quantities
    for z in range(4):
        gpz = gp_priors[z]

        logS = gpz["logSw4"][0] - 4 * gpz["logw0"][0]

        if len(gpz["logSw4"]) == 1:
            print(1)
            gp_priors[z]["logS"] = [logS]

        if len(gpz["logSw4"]) == 2:
            logS_var = gpz["logSw4"][1] ** 2 + 16 * gpz["logw0"][1] ** 2
            gp_priors[z]["logS"] = np.array([logS, np.sqrt(logS_var)])

    # # ############################
    # # ----- LIGHTCURVE FITTING -----
    # # ############################

    print("\nModeling Lightcurve...\n")

    # identify which quarters and seasons have data
    no_transits = (np.array(all_dtype) == "long_no_transits") + (
        np.array(all_dtype) == "short_no_transits"
    )
    which_quarters = np.sort(np.unique(np.hstack(transit_quarter)))
    which_quarters = which_quarters[
        ~np.isin(which_quarters, np.squeeze(np.argwhere(no_transits)))
    ]
    which_seasons = np.sort(np.unique(which_quarters % 4))

    # mean flux and jitter for each quarter
    nq = len(which_quarters)
    mbq = mean_by_quarter[which_quarters]
    vbq = var_by_quarter[which_quarters]

    # initialize alderaan.Ephemeris objects
    ephem = [None] * NPL
    for npl in range(NPL):
        ephem[npl] = Ephemeris(transit_inds[npl], regular_transit_times[npl])

    # initialize TransitParams objects
    theta = [None] * NPL
    for npl in range(NPL):
        theta[npl] = batman.TransitParams()
        theta[npl].per = ephem[npl].period
        theta[npl].t0 = 0.0  # t0 must be set to zero b/c we are warping TTVs
        theta[npl].rp = np.sqrt(DEPTHS[npl])
        theta[npl].b = 0.5
        theta[npl].T14 = DURS[npl]
        theta[npl].u = [U1, U2]
        theta[npl].limb_dark = "quadratic"

    # grab data
    t_ = [None] * nq
    f_ = [None] * nq
    e_ = [None] * nq
    m_ = [None] * nq  # m_ is "mask" not "model"

    for j, q in enumerate(which_quarters):
        m_[j] = all_mask[q].sum(axis=0) > 0
        t_[j] = all_time[q][m_[j]]
        f_[j] = all_flux[q][m_[j]]
        e_[j] = all_error[q][m_[j]]

    # build the GP kernel using a different noise model for each season
    kernel = [None] * 4
    for z in which_seasons:
        kernel[z] = GPterms.SHOTerm(
            S0=np.exp(gp_priors[z]["logS"][0]),
            w0=np.exp(gp_priors[z]["logw0"][0]),
            Q=np.exp(gp_priors[z]["logQ"][0]),
        )

    warped_t = []
    warped_x = []

    for npl in range(NPL):
        warped_t.append([])
        warped_x.append([])

        for j, q in enumerate(which_quarters):
            _warp_time, _warp_index = ephem[npl]._warp_times(t_[j], return_inds=True)
            _warp_legx = (_warp_index - transit_inds[npl][-1] // 2) / (
                transit_inds[npl][-1] / 2
            )

            warped_t[npl].append(_warp_time)
            warped_x[npl].append(_warp_legx)

    transit_legx = []

    for npl in range(NPL):
        transit_legx.append(centered_transit_inds[npl] / (transit_inds[npl][-1] / 2))

    phot_args = {}

    phot_args["time"] = t_
    phot_args["flux"] = f_
    phot_args["error"] = e_
    phot_args["quarters"] = which_quarters
    phot_args["warped_t"] = warped_t
    phot_args["warped_x"] = warped_x
    phot_args["exptime"] = texp
    phot_args["oversample"] = oversample
    phot_args["texp_offsets"] = texp_offsets

    ephem_args = {}
    ephem_args["transit_inds"] = centered_transit_inds
    ephem_args["transit_times"] = regular_transit_times
    ephem_args["transit_legx"] = transit_legx

    ncores = multipro.cpu_count() - 2
    ndim = 5 * NPL + 2
    logl = dynhelp.lnlike
    ptform = dynhelp.prior_transform
    logl_args = (NPL, theta, ephem_args, phot_args, [U1, U2], kernel)
    ptform_args = (NPL, DURS)
    chk_file = os.path.join(RESULTS_DIR, f"{TARGET}-dynesty.checkpoint")

    USE_MULTIPRO = False

    if USE_MULTIPRO:
        with dynesty.pool.Pool(
            ncores, logl, ptform, logl_args=logl_args, ptform_args=ptform_args
        ) as pool:
            sampler = dynesty.DynamicNestedSampler(
                pool.loglike,
                pool.prior_transform,
                ndim,
                bound="multi",
                sample="rwalk",
                pool=pool,
            )
            sampler.run_nested(
                checkpoint_file=chk_file, checkpoint_every=60, print_progress=VERBOSE
            )
            results = sampler.results

    else:
        sampler = dynesty.DynamicNestedSampler(
            logl,
            ptform,
            ndim,
            bound="multi",
            sample="rwalk",
            logl_args=logl_args,
            ptform_args=ptform_args,
        )
        sampler.run_nested(
            checkpoint_file=chk_file, checkpoint_every=60, print_progress=VERBOSE
        )
        results = sampler.results

    # ## Make diagnostic plots

    labels = []
    for npl in range(NPL):
        labels = labels + "C0_{0} C1_{0} r_{0} b_{0} T14_{0}".format(npl).split()

    labels = labels + "q1 q2".split()

    # run plot
    fig, ax = dyplot.runplot(results, logplot=True)
    plt.savefig(
        os.path.join(FIGURE_DIR, f"{TARGET}_dynesty_runplot.png"), bbox_inches="tight"
    )

    # trace plot
    for n in range(NPL):
        fig, ax = dyplot.traceplot(
            results,
            labels=labels[5 * n : 5 * (n + 1)],
            dims=np.arange(5 * n, 5 * (n + 1)),
        )
        plt.savefig(
            os.path.join(FIGURE_DIR, f"{TARGET}_dynesty_trace_{n:02d}.png"),
            bbox_inches="tight",
        )

    fig, ax = dyplot.traceplot(results, labels=labels[-2:], dims=np.array([-2, -1]))
    plt.savefig(
        os.path.join(FIGURE_DIR, f"{TARGET}_dynesty_trace_q1q2.png"),
        bbox_inches="tight",
    )

    # corner plot
    for n in range(NPL):
        fig, ax = dyplot.cornerplot(
            results,
            labels=labels[5 * n : 5 * (n + 1)],
            dims=np.arange(5 * n, 5 * (n + 1)),
        )
        plt.savefig(
            os.path.join(FIGURE_DIR, f"{TARGET}_dynesty_corner_{n:02d}.png"),
            bbox_inches="tight",
        )

    fig, ax = dyplot.cornerplot(results, labels=labels[-2:], dims=np.array([-2, -1]))
    plt.savefig(
        os.path.join(FIGURE_DIR, f"{TARGET}_dynesty_corner_q1q2.png"),
        bbox_inches="tight",
    )

    # ## Save sampling results

    hdu_list = io.dynesty_results_to_fits(results, PROJECT_DIR, MISSION, TARGET, RUN_ID)

    path = os.path.join(PROJECT_DIR, "Results", RUN_ID, TARGET)
    os.makedirs(path, exist_ok=True)
    hdu_list.writeto(os.path.join(path, f"{TARGET}-results.fits"), overwrite=True)

    # ## Exit program

    print("")
    print("+" * shutil.get_terminal_size().columns)
    print(
        f"Exoplanet recovery complete {datetime.now().strftime('%d-%b-%Y at %H:%M:%S')}"
    )
    print(f"Total runtime = {(timer()-global_start_time)/60:.1f} min")
    print("+" * shutil.get_terminal_size().columns)


if __name__ == "__main__":
    main()



__all__ = ["build_transit_model", "remove_known_transits", "inject_synthetic_transits"]


import aesara_theano_fallback.tensor as T
import numpy as np
import exoplanet as exo
import pymc3 as pm

from .constants import *



def build_transit_model(planets, data, limbdark, texp, oversample):
    """
    Docstring
    """
    npl = len(planets)
    
    tts  = [None]*npl
    inds = [None]*npl
    ror  = np.zeros(npl)
    b    = np.zeros(npl)
    dur  = np.zeros(npl)

    for n, p in enumerate(planets):
        tts[n]  = p.tts
        inds[n] = p.index
        ror[n] = np.sqrt(p.depth)
        b[n]   = p.impact
        dur[n] = p.duration
        
    starrystar = exo.LimbDarkLightCurve([limbdark[0],limbdark[1]])

    orbit = exo.orbits.TTVOrbit(transit_times=tts, 
                                transit_inds=inds, 
                                ror=ror,
                                b=b, 
                                duration=dur
                               )
    
    model_flux = [None]*len(data)
    
    for i, d in enumerate(data):
        light_curve = starrystar.get_light_curve(orbit=orbit, 
                                                 r=ror, 
                                                 t=d.time, 
                                                 oversample=oversample, 
                                                 texp=texp,
                                                )
        
        model_flux[i] = pm.math.sum(light_curve, axis=-1) + T.ones(len(d.time))
        
    return model_flux


def remove_known_transits(planets, data, limbdark, texp, oversample):
    """
    Docstring
    """
    model_flux = build_transit_model(planets, data, limbdark, texp, oversample)
    
    for i, d in enumerate(data):
        d.flux /= model_flux[i].eval()
        
    return data


def inject_synthetic_transits(planets, data, limbdark, texp, oversample):
    """
    Docstring
    """
    model_flux = build_transit_model(planets, data, limbdark, texp, oversample)
    
    for i, d in enumerate(data):
        d.flux *= model_flux[i].eval()
        
    return data


__all__ = [
    "make_chunklist",
    "build_sho_model",
    "generate_acf",
    "model_acf",
    "make_covariance_matrix",
    "generate_synthetic_noise",
    "make_gp_prior_dict",
]


import warnings

import aesara_theano_fallback.tensor as T
import astropy
from celerite2.theano import GaussianProcess
from celerite2.theano import terms as GPterms
import numpy as np
from numpy.linalg import LinAlgError
import pymc3 as pm
import scipy.signal as sig

from .constants import *
from .utils import boxcar_smooth, FFT_estimator


def make_chunklist(
    time,
    flux,
    cadno,
    Npts,
    sigma_reject=5.0,
    gap_tolerance=13,
    interpolate=True,
    cover=0.95,
):
    """
    Make an array of 'chunks' of data uninterupted by transits or data gaps

    Parameters
    ----------
    time : array-like
        out-of-transit time values (i.e. transits masked BEFORE passing array into make_chucklist)
    flux : array-like
        out-of-transit flux values
    cadno : array-like
        out-of-transit cadence numbers
    Npts : int
        number of points to use in each chunk; should be ~3x max transit duration
    sigma_reject : float
        sigma threshold for rejection of noisy chunks (default=5.0)
    gap_tolerance : int
        maximum number of consecutive missing cadences allowed (default=2)
    interpolate : bool
        True to perform linear interpolation with additive white noise over small gaps (default=True)
    cover : float between (0,1)
        fractional coverage required to consider a chunk 'good'

    Returns
    -------
    chunklist : list
        M x N list of data 'chunks' uninterrupted by transits or data gaps
    """
    chunklist = []

    i = 0
    loop = True
    while loop:
        # mark start and end cadence for this chunk
        cad_low = cadno[i]
        cad_high = cad_low + Npts + 1

        use = (cadno >= cad_low) * (cadno < cad_high)

        # check if there are gaps
        index = cadno[use] - cad_low
        gaps = np.hstack([1, index[1:] - index[:-1]])

        # make sure there are no long stretches of consecutive missing cadences
        if gaps.max() <= gap_tolerance:
            # pull time and flux chunks
            t_chunk = np.ones(Npts + 1) * 99  # use '99' to mark missing cadences
            f_chunk = np.ones(Npts + 1) * 99

            t_chunk[index] = time[use]
            f_chunk[index] = flux[use]

            # interpolate over missing cadences
            empty = t_chunk == 99

            t_chunk[empty] = np.interp(
                np.arange(Npts + 1)[empty], index, t_chunk[~empty]
            )
            f_chunk[empty] = np.interp(
                np.arange(Npts + 1)[empty], index, f_chunk[~empty]
            )
            f_chunk[empty] += np.random.normal(size=np.sum(empty)) * np.std(
                f_chunk[~empty]
            )

            # require at least X% coverage (default = 95%)
            if np.sum(~empty) / len(empty) > cover:
                chunklist.append(f_chunk)

            i += Npts

        # if there are missing cadences, move forward past any gaps
        else:
            i += np.argmax(gaps > gap_tolerance)

        # finish the loop if there is no more unusued flux
        if i > (len(time) - Npts - 1):
            loop = False

    # convert list to array
    chunklist = np.array(chunklist)

    # reject any chunks with unusually high medians or variability
    loop = True
    while loop:
        mad_chunk = astropy.stats.mad_std(chunklist, axis=1)
        med_chunk = np.median(chunklist, axis=1)

        bad = (
            np.abs(mad_chunk - np.median(mad_chunk)) / astropy.stats.mad_std(mad_chunk)
            > sigma_reject
        )
        bad += (
            np.abs(med_chunk - np.median(med_chunk)) / astropy.stats.mad_std(med_chunk)
            > sigma_reject
        )

        chunklist = chunklist[~bad]

        if np.sum(bad) == 0:
            loop = False

    return chunklist


def build_sho_model(t, y, var_method, fmin=None, fmax=None, f0=None, Q0=None):
    """
    Build PyMC3/celerite model for correlated noise using a single SHOTerm

    Parameters
    ----------
    t : array-like
        independent variable data (e.g. time)
    y : array-like
        corresponding dependent variable data (e.g. empirical ACF or flux)
    var_method : string
        automatic method for selecting y data variance
        'global' --> yvar = np.var(y)
        'local' --> yvar = np.var(y - local_trend)
        'fit' --> log_yvar is a free hyperparameter in the GP model
    fmin : float (optional)
        lower bound on (ordinary, not angular) frequency
    fmax : float (optional)
        upper bound on (ordinary, not angular) frequency
    f0 : float (optional)
        if provided, tight priors will be placed on this frequency
    Q0 : float (optional)
        if provided, Q will be fixed to this value

    Returns
    -------
    model : a PyMC3 model with a celerite SHOTERM
    """
    with pm.Model() as model:
        # frequency
        if fmin is None:
            fmin = 1 / (t.max() - t.min())
        if fmax is None:
            fmax = 2 / np.mean(np.diff(t))

        df = 1 / (t.max() - t.min())

        if f0 is None:
            logw0 = pm.Uniform(
                "logw0", lower=np.log(2 * pi * fmin), upper=np.log(2 * pi * fmax)
            )
        else:
            logw0 = pm.Bound(
                pm.Normal, lower=np.log(2 * pi * fmin), upper=np.log(2 * pi * fmax)
            )("logw0", mu=np.log(2 * pi * f0), sd=np.log((f0 + df) / f0))

        w0 = pm.Deterministic("w0", T.exp(logw0))

        # quality factor (downstream transit models accept logQ - easier to keep using it here)
        if Q0 is None:
            logQ_off = pm.Normal("logQ_off", mu=0, sd=5, testval=np.log(1e-3))
            logQ = pm.Deterministic("logQ", T.log(1 / np.sqrt(2) + T.exp(logQ_off)))
        else:
            logQ = pm.Deterministic("logQ", T.log(Q0))

        # amplitude parameter
        logSw4 = pm.Normal("logSw4", mu=np.log(np.var(y) * fmax**3), sd=10.0)
        S0 = pm.Deterministic("S0", T.exp(logSw4) / w0**4)

        # mean
        mean = pm.Normal("mean", mu=np.mean(y), sd=np.std(y))

        # diagonal variance
        if var_method == "global":
            yvar = np.var(y)
        elif var_method == "local":
            yvar = np.var(y - boxcar_smooth(y, 7))
        elif var_method == "fit":
            log_yvar = pm.Normal("log_yvar", mu=np.log(np.var(y)), sd=10.0)
            yvar = T.exp(log_yvar)
        else:
            raise ValueError("Must specify var_method as 'global', 'local', or 'fit'")

        # set up the GP
        kernel = GPterms.SHOTerm(S0=S0, w0=w0, Q=T.exp(logQ))
        gp = GaussianProcess(kernel, t=t, diag=yvar * T.ones(len(t)), mean=mean)
        gp.marginal("gp", observed=y)

        # track GP prediction
        pm.Deterministic("pred", gp.predict(y))

    return model


def generate_acf(time, flux, cadno, Npts, sigma_reject=5.0, keep_zero_lag=False):
    """
    Generate an autocorrelation function from a collection of out-of-transit data 'chunks'

    Parameters
    ----------
    time : array-like
        out-of-transit time values (i.e. transits masked BEFORE passing array into make_chucklist)
    flux : array-like
        out-of-transit flux values
    cadno : array-like
        out-of-transit cadence numbers
    Npts : int
        number of points to use in each chunk; should be ~3x max transit duration
    sigma_reject : float
        sigma threshold for rejection of noisy chunks (default=5.0)
    keep_zero_lag : bool
        default behavior (False) is to remove lag-zero term from ACF before returning

    Returns
    -------
    xcor : ndarray
        time-lag used to generate autocorrelation function with lag-zero value removed
    acor : ndarray
        autocorrelation function with lag-zero value removed
    """
    chunklist = make_chunklist(time, flux, cadno, Npts, sigma_reject=sigma_reject)

    Nsamples = chunklist.shape[0]

    # generate the autocorrelation function
    acor = np.zeros((Nsamples, 2 * Npts + 1))

    for i in range(Nsamples):
        chunk = chunklist[i] - np.median(chunklist[i])
        acor[i] = np.correlate(chunk, chunk, mode="full")

    acor = np.mean(acor, axis=0)
    acor = acor[Npts:] / acor[Npts]
    xcor = np.arange(Npts + 1)

    if keep_zero_lag:
        return xcor, acor
    else:
        return xcor[1:], acor[1:]


def model_acf(
    xcor,
    acor,
    fcut,
    fmin=None,
    fmax=None,
    crit_fap=0.003,
    method="smooth",
    window_length=None,
):
    """
    Model an empirical autocorrelation function (ACF) using one of several methods

    Parameters
    ----------
    xcor : array-like
        lag time values
    acor : array-like
        empirical autocorrelation function power at each time lag
    fcut : float
        cutoff value for seperating high vs. low frequencies
    fmin : float (optional)
        minimum frequency to check; if not provided this will be set to 1/baseline
    fmax : float (optional)
        maximum frequency to check; if not provided this will be set to the Nyquist frequency
    crit_fap : float
        critical false alarm probability for significant signals (default=0.003)
    method : string
        method to model low frequency component; either 'smooth' or 'savgol'  (default='smooth')
    window_length : int
        size of smoothing window  (required if method is 'smooth')

    Returns
    -------
    acor_emp, acor_mod, xf, yf, freqs
    """
    # arrays to hold empirical and model ACF
    acor_mod = np.zeros_like(acor)
    acor_emp = acor.copy()

    # 1st model component (low frequency)
    xf_L, yf_L, freqs_L, faps_L = FFT_estimator(
        xcor, acor_emp, fmin=fmin, fmax=fmax, crit_fap=crit_fap
    )

    low_freqs = freqs_L[freqs_L < fcut]

    # model the ACF with chosen method
    if method == "smooth":
        if window_length is None:
            raise ValueError("Must specify window_length if method == 'smooth'")

        acor_mirror = np.hstack([acor_emp[::-1], acor_emp])
        acor_mod = boxcar_smooth(acor_mirror, window_length)[len(acor_emp) :]

    elif method == "savgol":
        if window_length is None:
            if len(low_freqs) == 0:
                window_length = 45
            else:
                window_length = int((24 * 60) / low_freqs[0]) // 2
                window_length = window_length + (window_length % 2) + 1

        acor_mirror = np.hstack([acor_emp[::-1], acor_emp])
        acor_mod = sig.savgol_filter(
            acor_mirror, polyorder=2, window_length=window_length
        )[len(acor_emp) :]
        acor_mod = boxcar_smooth(acor_mod, 5)

    else:
        raise ValueError("method must be either 'smooth', or 'savgol'")

    # check for any high-frequency components
    xf_H, yf_H, freqs_H, faps_H = FFT_estimator(
        xcor, acor_emp - acor_mod, fmin=fmin, fmax=fmax, crit_fap=crit_fap, max_peaks=5
    )

    high_freqs = freqs_H[freqs_H > fcut]

    # combine freqs, return ACF and power spectrum
    freqs = np.hstack([low_freqs, high_freqs])

    return acor_emp, acor_mod, xf_L, yf_L, freqs


def make_covariance_matrix(acf, size=None):
    """
    Generate a square 2D covariance matrix from a 1D autocorrelation function

    Parameters
    ----------
    acf : array-like
        1D autocorrelation function not including lag-zero term (length N)
    size : int
        size of output covariance matrix (optional; if not given size --> N+1 x N+1)

    Returns
    -------
    covmatrix : ndarray
        n x n array; diagonal terms all equal 1.0
    """
    N = len(acf) + 1

    if size is None:
        n = N
    else:
        n = size

    if n > N:
        acf = np.hstack([acf, np.zeros(n - N)])

    covmatrix = np.zeros((n, n))

    for i in range(n):
        covmatrix[i, i + 1 :] = acf[: n - i - 1]

    covmatrix += covmatrix.swapaxes(0, 1)
    covmatrix += np.eye(n)

    return covmatrix


def generate_synthetic_noise(xcor, acor, n, sigma):
    """
    Generate synthetic correlated noise given a specified autorrelation function


    Parameters
    ----------
    xcor : array-like
        lag time values
    acor : array-like
        autocorrelation function power at each time lag
    n : int
        size of n x n covariance matrix
    sigma : float
        scale of white noise

    Returns
    -------
    x : ndarray
        1D array of time values (or some general independent coordinate)
    red_noise : ndarray
        synthetic correlated noise
    white_noise: ndarray
        gaussian noise vector used to generate red noise
    """
    # first make the covariance matrix
    covmatrix = make_covariance_matrix(acor, n)

    # decompose it
    try:
        L = np.linalg.cholesky(covmatrix)

    except LinAlgError:
        try:
            warnings.warn(
                "Covariance matrix not positive definite...adjusting automatically"
            )

            # decompose for eigenvalues and eigenvectors
            # matrix was constructed to be symmetric - if eigh doesn't work there is a serious problem
            eigenvalues, eigenvectors = np.linalg.eigh(covmatrix)

            # diagonalize
            D = np.diag(eigenvalues)

            # elementwise comparison with zero
            Z = np.zeros_like(D)
            Dz = np.maximum(D, Z)

            # generate a positive semi-definite matrix
            psdm = np.dot(np.dot(eigenvectors, Dz), eigenvectors.T)

            # now make it positive definite
            eps = np.min(np.abs(acor[acor != 0])) * 1e-6
            covmatrix = psdm + np.eye(n) * eps

            # renormalize
            covmatrix = covmatrix / covmatrix.max()

            # do Cholesky decomposition
            L = np.linalg.cholesky(covmatrix)

        except LinAlgError:
            warnings.warn(
                "Covariance matrix fatally broken...returning identity matrix"
            )
            covmatrix = np.eye(n)
            L = np.linalg.cholesky(covmatrix)

    # generate a vector of gaussian noise and remove any random covariance
    z = np.random.normal(size=n) * sigma

    # make correlated noise
    x = np.arange(n) * (xcor[1] - xcor[0])
    y = np.dot(L, z)

    return x, y - z, z


def make_gp_prior_dict(
    sho_trace, percs=[0.135, 2.275, 15.865, 50.0, 84.135, 97.725, 99.865]
):
    """
    Generates a list of percentiles from posteriors for each hyperparameter of a GP noise model
    The expected sho_trace should be the output of a PyMC3/Exoplanet model built with noise.build_sho_model()

    Assumes a specific set of input variable names from sho_trace:
      - ['logw0', 'logSw4', 'logQ']

    Parameters
    ----------
    sho_trace : PyMC3 multitrace
        trace output of a PyMC3/Exoplanet model built with noise.build_sho_model()
        will also work if sho_trace a dictionary with the necessary keys
    percs : list
        list of percentiles to return, by default 1- 2- 3- sigma and median

    Returns
    -------
    priors : dict
        Dictionary keys can be any combination of ['logw0', 'logSw4', 'logQ']
        Each key gives a list of values corresponding to specified percentiles from sho_trace
    """
    priors = {}
    priors["percentiles"] = percs

    try:
        varnames = sho_trace.varnames
    except Exception:
        varnames = list(sho_trace.keys())

    # get posterior percentiles of each hyperparameter and assign to a dictionary
    if np.isin("logw0", varnames):
        priors["logw0"] = np.percentile(sho_trace["logw0"], percs)

    if np.isin("logSw4", varnames):
        priors["logSw4"] = np.percentile(sho_trace["logSw4"], percs)

    if np.isin("logQ", varnames):
        priors["logQ"] = np.percentile(sho_trace["logQ"], percs)

    return priors



#######################################
# - Fit Transit Shape Independently - #
#######################################

# This script fits a transit lightcurve model for a single Kepler planet
# Overlapping transits in multiplanet systems are removed
# Transits are modeled using PyMC3/HMC following the umbrella sampling routine outlined in Gilbert 2022
# The code outputs a posterior MCMC chain for a *single* umbrella window

import os
import sys
import glob
import shutil
import warnings
from datetime import datetime
from timeit import default_timer as timer

print("")
print("+" * shutil.get_terminal_size().columns)
print("ALDERAAN Transit Fitting (single planet)")
print("Initialized {0}".format(datetime.now().strftime("%d-%b-%Y at %H:%M:%S")))
print("+" * shutil.get_terminal_size().columns)
print("")

# start program timer
global_start_time = timer()

# MCMC parameters are always manually set w/in the code
Nchain = 2
Ntune = 20000
Ndraw = 10000
target_accept = 0.95

# parse inputs
import argparse
import matplotlib as mpl

parser = argparse.ArgumentParser(
    description="Inputs for ALDERAAN transit fiting pipeline"
)
parser.add_argument(
    "--mission",
    default=None,
    type=str,
    required=True,
    help="Mission name; can be 'Kepler' or 'Simulated'",
)
parser.add_argument(
    "--target",
    default=None,
    type=str,
    required=True,
    help="Target name; format should be K00000 or S00000",
)
parser.add_argument(
    "--planet_no",
    default=None,
    type=int,
    required=True,
    help="ALDERAAN zero-indexed planet identifier (i.e. *NOT* KOI_ID)",
)
parser.add_argument(
    "--root_dir",
    default=None,
    type=str,
    required=True,
    help="Root directory for system",
)
parser.add_argument(
    "--project_dir",
    default=None,
    type=str,
    required=True,
    help="Project directory for accessing lightcurve data and saving outputs",
)
parser.add_argument(
    "--catalog",
    default=None,
    type=str,
    required=True,
    help="CSV file containing input planetary parameters",
)
parser.add_argument(
    "--umbrella",
    default=None,
    type=str,
    required=True,
    help="Umbrella can be 'N', 'T', or 'G'",
)
parser.add_argument(
    "--interactive",
    default=False,
    type=bool,
    required=False,
    help="'True' to enable interactive plotting; by default matplotlib backend will be set to 'Agg'",
)

args = parser.parse_args()
MISSION = args.mission
TARGET = args.target
PLANET_NO = args.planet_no
ROOT_DIR = args.root_dir
PROJECT_DIR = ROOT_DIR + args.project_dir
CATALOG = PROJECT_DIR + "Catalogs/" + args.catalog
UMBRELLA = args.umbrella

# set plotting backend
if args.interactive == False:
    mpl.use("agg")

# set environment variables
sys.path.append(PROJECT_DIR)

# build directory structure
FIGURE_DIR = PROJECT_DIR + "Figures/" + TARGET + "/"
TRACE_DIR = PROJECT_DIR + "Traces/" + TARGET + "/"
QUICK_TTV_DIR = PROJECT_DIR + "QuickTTVs/" + TARGET + "/"
DLC_DIR = PROJECT_DIR + "Detrended_lightcurves/" + TARGET + "/"
NOISE_DIR = PROJECT_DIR + "Noise_models/" + TARGET + "/"

# check if all the output directories exist and if not, create them
if os.path.exists(FIGURE_DIR) == False:
    os.mkdir(FIGURE_DIR)

if os.path.exists(TRACE_DIR) == False:
    os.mkdir(TRACE_DIR)

if os.path.exists(QUICK_TTV_DIR) == False:
    os.mkdir(QUICK_TTV_DIR)

if os.path.exists(DLC_DIR) == False:
    os.mkdir(DLC_DIR)

if os.path.exists(NOISE_DIR) == False:
    os.mkdir(NOISE_DIR)


# import packages
import astropy.stats
from astropy.io import fits
import json
import matplotlib.pyplot as plt
import numpy as np
import numpy.polynomial.polynomial as poly
import pandas as pd
from scipy import stats

import pymc3 as pm
import pymc3_ext as pmx
import exoplanet as exo
import aesara_theano_fallback.tensor as T
from aesara_theano_fallback import aesara as theano
from celerite2.theano import GaussianProcess
from celerite2.theano import terms as GPterms

from alderaan.constants import *
import alderaan.io as io
from alderaan.detrend import make_transitmask
from alderaan.LiteCurve import LiteCurve
from alderaan.Planet import Planet


# flush buffer to avoid mixed outputs from progressbar
sys.stdout.flush()

# turn off FutureWarnings
warnings.filterwarnings("ignore", category=FutureWarning)

# check for interactive matplotlib backends
if np.any(np.array(["agg", "png", "svg", "pdf", "ps"]) == mpl.get_backend()):
    iplot = False
else:
    iplot = True

# echo theano cache directory
print("theano cache: {0}\n".format(theano.config.compiledir))


################
# - DATA I/O - #
################

print("\nLoading data...\n")

# Read in planet and star properties from csv file
target_dict = pd.read_csv(CATALOG)

# set KOI_ID global variable
if MISSION == "Kepler":
    KOI_ID = TARGET
elif MISSION == "Simulated":
    KOI_ID = "K" + TARGET[1:]
else:
    raise ValueError("MISSION must be 'Kepler' or 'Simulated'")

# pull relevant quantities and establish GLOBAL variables
use = np.array(target_dict["koi_id"]) == KOI_ID

KIC = np.array(target_dict["kic_id"], dtype="int")[use]
NPL = np.array(target_dict["npl"], dtype="int")[use]

PERIODS = np.array(target_dict["period"], dtype="float")[use]
DEPTHS = np.array(target_dict["depth"], dtype="float")[use] * 1e-6  # [ppm] --> []
DURS = np.array(target_dict["duration"], dtype="float")[use] / 24  # [hrs] --> [days]

# do some consistency checks
if all(k == KIC[0] for k in KIC):
    KIC = KIC[0]
else:
    raise ValueError("There are inconsistencies with KIC in the csv input file")

if all(n == NPL[0] for n in NPL):
    NPL = NPL[0]
else:
    raise ValueError("There are inconsistencies with NPL in the csv input file")

# sort planet parameters by period
order = np.argsort(PERIODS)

PERIODS = PERIODS[order]
DEPTHS = DEPTHS[order]
DURS = DURS[order]


# Read in filtered lightcurves
# These can be generated by running the script "analyze_autocorrelated_noise.py"

try:
    lc = io.load_detrended_lightcurve(DLC_DIR + TARGET + "_lc_filtered.fits")
    lc.season = lc.quarter % 4
except:
    lc = None

try:
    sc = io.load_detrended_lightcurve(DLC_DIR + TARGET + "_sc_filtered.fits")
    sc.season = sc.quarter % 4
except:
    sc = None


# Read in quick transit times
# These can be generated by running the script "detrend_and_estimate_ttvs.py"

epochs = np.zeros(NPL)
periods = np.zeros(NPL)
ephemeris = [None] * NPL

transit_inds = []
indep_transit_times = []
quick_transit_times = []

for npl in range(NPL):
    fname_in = QUICK_TTV_DIR + TARGET + "_{:02d}".format(npl) + "_quick.ttvs"
    data_in = np.genfromtxt(fname_in)

    transit_inds.append(np.array(data_in[:, 0], dtype="int"))
    indep_transit_times.append(np.array(data_in[:, 1], dtype="float"))
    quick_transit_times.append(np.array(data_in[:, 2], dtype="float"))

    # do a quick fit to get a linear ephemeris
    pfit = poly.polyfit(transit_inds[npl], quick_transit_times[npl], 1)

    epochs[npl] = pfit[0]
    periods[npl] = pfit[1]
    ephemeris[npl] = poly.polyval(transit_inds[npl], pfit)

# make sure transit_inds are zero-indexed
for npl in range(NPL):
    transit_inds[npl] = np.array(transit_inds[npl] - transit_inds[npl][0], dtype="int")


# Make OMC plots
fig, axes = plt.subplots(NPL, figsize=(12, 3 * NPL))
if NPL == 1:
    axes = [axes]

for npl in range(NPL):
    xtime = ephemeris[npl]
    yomc_i = (indep_transit_times[npl] - ephemeris[npl]) * 24 * 60
    yomc_q = (quick_transit_times[npl] - ephemeris[npl]) * 24 * 60

    axes[npl].plot(xtime, yomc_i, "o", c="lightgrey")
    axes[npl].plot(xtime, yomc_q, lw=2, c="C{0}".format(npl))
    axes[npl].set_ylabel("O-C [min]", fontsize=20)
axes[NPL - 1].set_xlabel("Time [BJKD]", fontsize=20)

if iplot:
    plt.show()
else:
    plt.close()


#####################
# - PRELIMINARIES - #
#####################

print("\nRunning preliminaries...\n")

# Establish time baseline
time_min = []
time_max = []

if sc is not None:
    time_min.append(sc.time.min())
    time_max.append(sc.time.max())

if lc is not None:
    time_min.append(lc.time.min())
    time_max.append(lc.time.max())

TIME_START = np.min(time_min)
TIME_END = np.max(time_max)


# Put epochs in range (TIME_START, TIME_START + PERIOD)
for npl in range(NPL):
    if epochs[npl] < TIME_START:
        adj = 1 + (TIME_START - epochs[npl]) // periods[npl]
        epochs[npl] += adj * periods[npl]

    if epochs[npl] > (TIME_START + periods[npl]):
        adj = (epochs[npl] - TIME_START) // periods[npl]
        epochs[npl] -= adj * periods[npl]


# Identify and remove overlapping transits
overlap = []

for i in range(NPL):
    overlap.append(np.zeros(len(ephemeris[i]), dtype="bool"))

    for j in range(NPL):
        if i != j:
            for tt in ephemeris[j]:
                overlap[i] += np.abs(ephemeris[i] - tt) < (DURS[i] + DURS[j] + lcit)

ephemeris = [ephemeris[npl][~overlap[npl]] for npl in range(NPL)]
transit_inds = [transit_inds[npl][~overlap[npl]] for npl in range(NPL)]
quick_transit_times = [quick_transit_times[npl][~overlap[npl]] for npl in range(NPL)]


# Track which quarter each transit falls in
if lc is not None:
    lc_quarters = np.unique(lc.quarter)
else:
    lc_quarters = np.array([])

if sc is not None:
    sc_quarters = np.unique(sc.quarter)
else:
    sc_quarters = np.array([])

quarters = np.sort(np.hstack([lc_quarters, sc_quarters]))
seasons = np.sort(np.unique(quarters % 4))

# get list of threshold times between quarters
thresh = np.zeros(len(quarters) + 1)
thresh[0] = TIME_START

for j, q in enumerate(quarters):
    if np.isin(q, sc_quarters):
        thresh[j + 1] = sc.time[sc.quarter == q].max()
    if np.isin(q, lc_quarters):
        thresh[j + 1] = lc.time[lc.quarter == q].max()

thresh[0] -= 1.0
thresh[-1] += 1.0

# track individual transits
transit_quarter = [None] * NPL

for npl in range(NPL):
    tts = ephemeris[npl]
    transit_quarter[npl] = np.zeros(len(tts), dtype="int")

    for j, q in enumerate(quarters):
        transit_quarter[npl][(tts >= thresh[j]) * (tts < thresh[j + 1])] = q


# Make transit masks
if sc is not None:
    sc_mask = np.zeros((NPL, len(sc.time)), dtype="bool")
    for npl in range(NPL):
        sc_mask[npl] = make_transitmask(sc.time, quick_transit_times[npl], masksize=1.5)

if lc is not None:
    lc_mask = np.zeros((NPL, len(lc.time)), dtype="bool")
    for npl in range(NPL):
        lc_mask[npl] = make_transitmask(lc.time, quick_transit_times[npl], masksize=1.5)


# Grab data near transits for each quarter
all_time = [None] * 18
all_flux = [None] * 18
all_error = [None] * 18
all_mask = [None] * 18
all_dtype = ["none"] * 18

for q in range(18):
    if sc is not None:
        if np.isin(q, sc.quarter):
            use = (sc_mask.sum(0) != 0) * (sc.quarter == q)

            if np.sum(use) > 45:
                all_time[q] = sc.time[use]
                all_flux[q] = sc.flux[use]
                all_error[q] = sc.error[use]
                all_mask[q] = sc_mask[:, use]
                all_dtype[q] = "short"
            else:
                all_dtype[q] = "short_no_transits"

    if lc is not None:
        if np.isin(q, lc.quarter):
            use = (lc_mask.sum(0) != 0) * (lc.quarter == q)

            if np.sum(use) > 5:
                all_time[q] = lc.time[use]
                all_flux[q] = lc.flux[use]
                all_error[q] = lc.error[use]
                all_mask[q] = lc_mask[:, use]
                all_dtype[q] = "long"
            else:
                all_dtype[q] = "long_no_transits"


# set oversampling factors and expoure times
oversample = np.zeros(18, dtype="int")
texp = np.zeros(18)

oversample[np.array(all_dtype) == "short"] = 1
oversample[np.array(all_dtype) == "long"] = 15

texp[np.array(all_dtype) == "short"] = scit
texp[np.array(all_dtype) == "long"] = lcit


# track mean and variance of each quarter
mean_by_quarter = np.ones(18) * np.nan
var_by_quarter = np.ones(18) * np.nan

for q in range(18):
    if sc is not None:
        if np.isin(q, sc.quarter):
            mean_by_quarter[q] = np.mean(sc.flux[sc.quarter == q])
            var_by_quarter[q] = np.var(sc.flux[sc.quarter == q])

    if lc is not None:
        if np.isin(q, lc.quarter):
            mean_by_quarter[q] = np.mean(lc.flux[lc.quarter == q])
            var_by_quarter[q] = np.var(lc.flux[lc.quarter == q])


# Use Legendre polynomials over transit times for better orthogonality; "x" is in the range (-1,1)
# The current version of the code only uses 1st order polynomials, but 2nd and 3rd are retained for posterity
Leg0 = []
Leg1 = []
Leg2 = []
Leg3 = []
t = []

# this assumes a baseline in the range (TIME_START,TIME_END)
for npl in range(NPL):
    t.append(ephemeris[npl])
    x = 2 * (t[npl] - TIME_START) / (TIME_END - TIME_START) - 1

    Leg0.append(np.ones_like(x))
    Leg1.append(x.copy())
    Leg2.append(0.5 * (3 * x**2 - 1))
    Leg3.append(0.5 * (5 * x**3 - 3 * x))


# Read in GP noise model priors from analyze_autocorrelated_noise.py
gp_percs = []

for z in range(4):
    try:
        fname_in = NOISE_DIR + TARGET + "_shoterm_gp_priors_{0}.txt".format(z)

        with open(fname_in) as infile:
            gp_percs.append(json.load(infile))

    except:
        gp_percs.append(None)

# convert the percentile priors into Gaussians
gp_priors = []

for z in range(4):
    if gp_percs[z] is not None:
        gpz = {}

        for k in gp_percs[z].keys():
            if k != "percentiles":
                perc = np.array(gp_percs[z]["percentiles"])

                med = np.array(gp_percs[z][k])[perc == 50.0][0]
                err1 = np.array(gp_percs[z][k])[perc == 84.135][0]
                err2 = np.array(gp_percs[z][k])[perc == 15.865][0]

                dev = np.sqrt((err1 - med) ** 2 / 2 + (err2 - med) ** 2 / 2)

                gpz[k] = (med, dev)

        gp_priors.append(gpz)

    else:
        # these are dummy values that effectively create a zero-amplitude kernel
        gpz = {}
        gpz["logw0"] = [np.log(2 * pi / (7 * DURS.max()))]
        gpz["logSw4"] = [-100.0]
        gpz["logQ"] = [np.log(1 / np.sqrt(2))]

        gp_priors.append(gpz)

# calculate a few convenience quantities
for z in range(4):
    gpz = gp_priors[z]

    logS = gpz["logSw4"][0] - 4 * gpz["logw0"][0]

    if len(gpz["logSw4"]) == 1:
        gp_priors[z]["logS"] = [logS]

    if len(gpz["logSw4"]) == 2:
        logS_var = gpz["logSw4"][1] ** 2 + 16 * gpz["logw0"][1] ** 2
        gp_priors[z]["logS"] = np.array([logS, np.sqrt(logS_var)])


##########################
# - LIGHTCURVE FITTING - #
##########################

print("\nModeling Lightcurve...\n")

for npl in range(NPL):
    if npl == PLANET_NO:
        print("PLANET", npl)

        with pm.Model() as shape_model:
            # identify which quarters and seasons have data
            which_quarters = np.unique(transit_quarter[npl])
            which_seasons = np.unique(which_quarters % 4)

            # limb darkening
            u = exo.distributions.QuadLimbDark("LD_U", testval=[0.40, 0.25])

            # draw (r,b) parameters
            rmin, rmax = 1e-5, 0.99

            if UMBRELLA == "N":
                # draw (log_r, b) --> (r,b)
                log_r = pm.Uniform(
                    "LN_ROR",
                    lower=np.log(rmin),
                    upper=np.log(rmax),
                    testval=np.log(np.sqrt(DEPTHS[npl])),
                )
                r = pm.Deterministic("ROR", T.exp(log_r))
                b = pm.Uniform("IMPACT", lower=0, upper=1 - r)
                g = pm.Deterministic("GAMMA", (1 - b) / r)

                # this adjustment term makes samples uniform in the (r,b) plane
                adj = pm.Potential("adj", T.log(1 - r) + T.log(r))

                # umbrella bias
                norm = 1 / rmin - 1.5
                psi = pm.Potential(
                    "PSI", T.log(T.switch(T.lt(g, 2), g - 1, 1.0)) / norm
                )

            elif UMBRELLA == "T":
                # draw (log_r, gamma) --> (r,b)
                log_r = pm.Uniform(
                    "LN_ROR",
                    lower=np.log(rmin),
                    upper=np.log(rmax),
                    testval=np.log(np.sqrt(DEPTHS[npl])),
                )
                r = pm.Deterministic("ROR", T.exp(log_r))
                g = pm.Uniform(
                    "GAMMA", lower=0, upper=T.switch(r < 0.5, 2, 1 / r), testval=1.0
                )
                b = pm.Deterministic("IMPACT", 1 - g * r)

                # Jacobian for (r,b) --> (r,gamma)
                jac = pm.Potential("jac", T.log(1 / r))

                # this adjustment term makes samples uniform in the (r,b) plane
                adj = pm.Potential(
                    "adj", 2 * T.log(r) + T.switch(r < 0.5, T.log(2 * r), 0.0)
                )

                # umbrella bias
                norm = 1.0
                psi = pm.Potential("PSI", T.log(T.switch(T.lt(g, 1), g, 2 - g)) / norm)

            elif UMBRELLA == "G":
                # draw (log_lambda, gamma) --> (r,b)
                g = pm.Uniform("GAMMA", lower=-0.99, upper=1.0, testval=0.0)
                log_lam = pm.Uniform(
                    "LN_LAM",
                    lower=np.log((g + 1) * rmin**2),
                    upper=np.log((g + 1) * rmax**2),
                )
                lam = pm.Deterministic("LAM", T.exp(log_lam))

                r = pm.Deterministic("ROR", pm.math.sqrt(lam / (g + 1)))
                log_r = pm.Deterministic("LN_ROR", T.log(r))
                b = pm.Deterministic("IMPACT", 1 - g * r)

                # Jacobian for (r,b) --> (lambda,gamma)
                jac = pm.Potential("jac", T.log(2 + 2 * g))

                # this adjustment term makes samples uniform in the (r,b) plane
                adj = pm.Potential("adj", -T.log(2 + 2 * g) + 2 * T.log(r))

                # umbrella bias
                norm = 1.0
                psi = pm.Potential(
                    "PSI", T.log(T.switch(T.lt(g, 0), 1 + g, 1 - g)) / norm
                )

            else:
                raise ValueError("Umbrella must be 'N', 'T', or 'G'")

            # enforce log-uniform prior on r
            r_marginal = pm.Potential("r_marginal", -T.log(1 + r) - T.log(r))

            # draw transit duration
            log_dur = pm.Uniform(
                "LN_DUR14",
                lower=np.log(scit),
                upper=np.log(3 * DURS[npl]),
                testval=np.log(DURS[npl]),
            )
            dur = pm.Deterministic("DUR14", T.exp(log_dur))

            # polynomial TTV parameters (coefficients for Legendre polynomials)
            C0 = pm.Normal(
                "C0", mu=0, sd=3 * np.std(quick_transit_times[npl] - ephemeris[npl])
            )
            C1 = pm.Normal(
                "C1", mu=0, sd=3 * np.std(quick_transit_times[npl] - ephemeris[npl])
            )

            # transit times
            transit_times = pm.Deterministic(
                "TTIMES", quick_transit_times[npl] + C0 * Leg0[npl] + C1 * Leg1[npl]
            )

            # set up limb darkened star and planetary orbit
            starrystar = exo.LimbDarkLightCurve(u)
            orbit = exo.orbits.TTVOrbit(
                transit_times=[transit_times],
                transit_inds=[transit_inds[npl]],
                b=b,
                duration=dur,
                ror=r,
            )

            # track period, epoch, and stellar density
            P = pm.Deterministic("PERIOD", orbit.period)
            T0 = pm.Deterministic("T0", orbit.t0)

            # build the GP kernel using a different noise model for each season
            logS = [None] * 4
            logw0 = [None] * 4
            logQ = [None] * 4
            kernel = [None] * 4

            for z in which_seasons:
                gpz = gp_priors[z]

                # logS[z]   = pm.Normal('GP_LNS_{0}'.format(z), mu=gpz['logS'][0], sd=gpz['logS'][1])
                # logw0[z]  = pm.Normal('GP_LNOM_{0}'.format(z), mu=gpz['logw0'][0], sd=gpz['logw0'][1])
                logS[z] = gpz["logS"][0]
                logw0[z] = gpz["logw0"][0]
                logQ[z] = gpz["logQ"][0]
                kernel[z] = GPterms.SHOTerm(
                    S0=T.exp(logS[z]), w0=T.exp(logw0[z]), Q=T.exp(logQ[z])
                )

            # mean flux and jitter for each quarter
            nq = len(which_quarters)
            mbq = mean_by_quarter[which_quarters]
            vbq = var_by_quarter[which_quarters]

            flux0 = pm.Normal("FLUXZPT", mu=mbq, sd=np.sqrt(vbq), shape=nq)
            log_jit = pm.Normal(
                "LN_JIT", mu=np.log(vbq / 10), sd=5.0 * np.ones(nq), shape=nq
            )

            # now evaluate the model
            light_curves = [None] * nq
            model_flux = [None] * nq
            flux_err = [None] * nq
            gp = [None] * nq

            for j, q in enumerate(which_quarters):
                # grab time and flux
                t_ = all_time[q][all_mask[q][npl]]
                f_ = all_flux[q][all_mask[q][npl]]

                # calculate light curves
                light_curves[j] = starrystar.get_light_curve(
                    orbit=orbit, r=r, t=t_, oversample=oversample[q], texp=texp[q]
                )

                model_flux[j] = pm.math.sum(light_curves[j], axis=-1) + flux0[
                    j
                ] * T.ones(len(t_))
                flux_err[j] = T.sqrt(
                    np.mean(all_error[q]) ** 2 + T.exp(log_jit[j])
                ) / np.sqrt(2)

                # here's the GP (w/ kernel by season)
                gp[j] = GaussianProcess(
                    kernel[q % 4],
                    t=t_,
                    diag=flux_err[j] ** 2 * T.ones(len(t_)),
                    mean=model_flux[j],
                )

                gp[j].marginal("gp_{0}".format(j), observed=f_)

        with shape_model:
            shape_map = shape_model.test_point
            shape_map = pmx.optimize(start=shape_map, vars=[flux0, log_jit])
            shape_map = pmx.optimize(start=shape_map, vars=[C0, C1])
            shape_map = pmx.optimize(start=shape_map, vars=[r, b, dur])

        with shape_model:
            trace = pmx.sample(
                tune=Ntune,
                draws=Ndraw,
                start=shape_map,
                chains=Nchain,
                target_accept=target_accept,
                return_inferencedata=True,
            )

            summary = pm.summary(trace)


# Write MCMC trace to .fits file
print("\nWriting MCMC trace to .fits\n")

# set up header info
summary_info = [
    # ['STATISTIC', 'DESCRIPTION'],
    ["MEAN", "Mean of the posterior"],
    ["SD", "Standard deviation of the posterior"],
    ["HDI_03", "Highest density interval of posterior at 3%"],
    ["HDI_97", "Highest density interval of posterior at 97%"],
    ["MCSE_MU", "Markov Chain Standard Error of posterior mean"],
    ["MCSE_SD", "Markov Chain Standard Error of posterior std dev"],
    ["ESS_BULK", "Effective sample size for bulk posterior"],
    ["ESS_TAIL", "Effective sample size for tail posterior"],
    ["R_HAT", "Estimate of rank normalized R-hat statistic"],
]

data_info = [
    # ['VARIABLE', 'UNITS',         'DESCRIPTION',   'PRIOR'],
    ["PERIOD", "days", "Orbital period", "Normal"],
    ["T0", "days", "Mid-point of first transit", "Normal"],
    [
        "LN_ROR",
        "dimensionless",
        "Natural log of planet-to-star radius ratio",
        "Uniform",
    ],
    ["LN_DUR14", "days", "Natural log of transit duration: 1st-4th contact", "Uniform"],
    ["IMPACT", "dimensionless", "Transit impact parameter", "Uniform"],
    [
        "LD_U1",
        "dimensionless",
        "1st quadratic stellar limb darkening coefficient",
        "Kipping (2013)",
    ],
    [
        "LD_U2",
        "dimensionless",
        "2nd quadratic stellar limb darkening coefficient",
        "Kipping (2013)",
    ],
    ["FLUXZPT", "dimensionless", "Mean flux offset", "Normal"],
    ["LN_JIT", "dimensionless", "Photometric jitter, added in quadrature", "Normal"],
    ["PSI", "dimensionless", "Umbrella bias", "Gilbert (2022)"],
    ["GAMMA", "dimensionless", "Grazing coordinate", "Gilbert (2022)"],
    ["LN_LIKE", "dimensionless", "Natural log-likelihood of the model fit", ""],
    ["QUALITY", "bool", "Data quality flag indicating divergent samples", ""],
]

SUMMARY_DTYPE = [
    ("PARAM_NAME", "U10"),
    ("MEAN", "<f8"),
    ("SD", "<f8"),
    ("HDI_03", "<f8"),
    ("HDI_97", "<f8"),
    ("MCSE_MU", "<f8"),
    ("MCSE_SD", "<f8"),
    ("ESS_BULK", "<f8"),
    ("ESS_TAIL", "<f8"),
    ("R_HAT", "<f8"),
]


# Build trace dataframe
trace_df = pd.DataFrame()

# split 2-dim LD_U into separate columns
for k in list(trace.posterior.keys()):
    if k == "LD_U":
        trace_df["LD_U1"] = list(trace.posterior["LD_U"][:, :, 0].data.reshape(-1))
        trace_df["LD_U2"] = list(trace.posterior["LD_U"][:, :, 1].data.reshape(-1))

    elif k == "FLUXZPT":
        for j, q in enumerate(which_quarters):
            trace_df["FLUXZPT_{0}".format(str(q).zfill(2))] = list(
                trace.posterior["FLUXZPT"][:, :, j].data.reshape(-1)
            )

    elif k == "LN_JIT":
        for j, q in enumerate(which_quarters):
            trace_df["LN_JIT_{0}".format(str(q).zfill(2))] = list(
                trace.posterior["LN_JIT"][:, :, j].data.reshape(-1)
            )

    elif k == "TTIMES":
        pass

    else:
        trace_df[k] = trace.posterior[k].data.reshape(-1)

# pull likelihood
ln_like = np.zeros((Nchain, Ndraw), dtype="float")
for dv in trace.log_likelihood.data_vars:
    ln_like += np.squeeze(trace.log_likelihood[dv].data)

trace_df["LN_LIKE"] = ln_like.reshape(-1)

# pull quality flags
trace_df["QUALITY"] = trace["sample_stats"]["diverging"].data.reshape(-1).astype(int)

# to avoid redundancy, prefer LN_ROR over ROR
if "ROR" in list(summary.index) and "LN_ROR" in list(summary.index):
    trace_df.drop("ROR", axis=1, inplace=True)

# to avoid redundancy, prefer LN_DUR14 over DUR14
if "DUR14" in list(summary.index) and "LN_DUR14" in list(summary.index):
    trace_df.drop("DUR14", axis=1, inplace=True)


# Build summary dataframe
summary_df = pd.DataFrame(columns=summary.keys())


def mapper_fluxzpt(index):
    return index[:7] + "_{0}".format(str(index[8:-1].zfill(2)))


def mapper_ln_jit(index):
    return index[:6] + "_{0}".format(str(index[7:-1].zfill(2)))


def mapper_ld_u(index):
    return index[:4] + "{0}".format(int(index[5]) + 1)


for idx in summary.index:
    if np.isin(idx, trace_df.keys()):
        summary_df = pd.concat([summary_df, summary[summary.index == idx]])

    elif idx[:7] == "FLUXZPT":
        summary_df = pd.concat([summary_df, summary[summary.index == idx]])
        summary_df.rename(index={idx: mapper_fluxzpt(idx)}, inplace=True)

    elif idx[:6] == "LN_JIT":
        summary_df = pd.concat([summary_df, summary[summary.index == idx]])
        summary_df.rename(index={idx: mapper_ln_jit(idx)}, inplace=True)

    elif idx[:4] == "LD_U":
        summary_df = pd.concat([summary_df, summary[summary.index == idx]])
        summary_df.rename(index={idx: mapper_ld_u(idx)}, inplace=True)

    elif idx[:6] == "PERIOD":
        summary_df = pd.concat([summary_df, summary[summary.index == idx]])
        summary_df.rename(index={idx: "PERIOD"}, inplace=True)

    elif idx[:2] == "T0":
        summary_df = pd.concat([summary_df, summary[summary.index == idx]])
        summary_df.rename(index={idx: "T0"}, inplace=True)


# Build primary HDU and header
timestamp = datetime.strptime(
    trace.posterior.attrs["created_at"], "%Y-%m-%dT%H:%M:%S.%f"
)

hdu = fits.PrimaryHDU()
hdu.header["TARGET"] = (TARGET + "-" + str(PLANET_NO).zfill(2), "Target name")
hdu.header["PHOTOSRC"] = (MISSION, "Source of photometry")
hdu.header["UMBRELLA"] = (UMBRELLA, "see Gilbert 2022")
hdu.header["DATETIME"] = (
    timestamp.strftime("%Y-%m-%d %H:%M:%S"),
    "Date and time of analysis",
)
hdu.header["COMMENT"] = "ALDERAAN single planet transit fit"

f_name = PROJECT_DIR + "Traces/{0}/{0}-{1}_{2}_indep.fits".format(
    TARGET, str(PLANET_NO).zfill(2), UMBRELLA
)
hdu.writeto(f_name, overwrite=True)

# store data from dataframes into a single FITS file
with fits.open(f_name) as hduL:
    # Store trace data and summary data to HDUs for each model
    rec_data = trace_df.to_records(index=False)
    hdu_data = fits.BinTableHDU(data=rec_data, name="POSTERIOR")

    rec_summ = summary_df.to_records().astype(SUMMARY_DTYPE)
    hdu_summ = fits.BinTableHDU(data=rec_summ, name="MCMC_STATS")

    hduL.append(hdu_data)
    hduL.append(hdu_summ)

    # posterior data
    data_head = hduL["POSTERIOR"].header
    data_head["RUNTIME"] = (
        np.round(trace.posterior.attrs["sampling_time"], 2),
        "Duration of model sampling (seconds)",
    )
    data_head["NTUNE"] = (Ntune, "Number of tuning steps")
    data_head["NDRAW"] = (Ndraw, "Number of sampler draws")
    data_head["NCHAIN"] = (Nchain, "Number of sampler chains")

    for i, dinf in enumerate(data_info):
        data_head[dinf[0]] = (dinf[1], dinf[2])

    # MCMC sampler statistics
    summ_head = hduL["MCMC_STATS"].header

    for i, sinf in enumerate(summary_info):
        summ_head[sinf[0]] = ("", sinf[1])

    hduL.writeto(f_name, overwrite=True)


# Exit program

print("")
print("+" * shutil.get_terminal_size().columns)
print(
    "Exoplanet recovery complete {0}".format(
        datetime.now().strftime("%d-%b-%Y at %H:%M:%S")
    )
)
print("Total runtime = %.1f min" % ((timer() - global_start_time) / 60))
print("+" * shutil.get_terminal_size().columns)



__all__ = ['LiteCurve']

from astropy.io import fits
import glob
import lightkurve as lk
import numpy as np
from alderaan.constants import kepler_lcit, kepler_scit


class LiteCurve:
    """LiteCurve
    """
    def __init__(self, *args, **kwargs):
        
        if len(args) == 0:
            self = self._set_empty_attribute_arrays()
        
        elif len(args) == 1 and isinstance(args[0], list):
            if all([isinstance(lc, LiteCurve) for lc in args[0]]):
                self = self.from_list(*args, **kwargs)
            else:
                raise TypeError("Unexpected input types in list")
        
        elif (len(args) > 1) and isinstance(args[0], str):
            if 'data_source' not in kwargs:
                raise ValueError("Missing required keyword argmument 'data_source")
            else:
                data_source = kwargs.pop('data_source')

            if data_source == 'Kepler PDCSAP':
                self = self.from_kplr_pdcsap(*args, **kwargs)
            elif data_source == 'ALDERAAN':
                self = self.from_alderaan(*args, **kwargs)
            else:
                raise ValueError(f"Unsupported data_source: {data_source}")      
        
        else:
            raise TypeError("Unsupported init signature")
    
    
    @classmethod
    def _set_empty_attribute_arrays(cls):
        lc_instance = cls.__new__(cls)
        lc_instance.time = np.array([]).astype(float)
        lc_instance.flux = np.array([]).astype(float)
        lc_instance.error = np.array([]).astype(float)
        lc_instance.cadno = np.array([]).astype(int)
        lc_instance.visit = np.array([]).astype(int)
        lc_instance.obsmode = np.array([]).astype(str)
        lc_instance.quality = np.array([]).astype(bool)
        return lc_instance

    
    @classmethod
    def from_list(cls, litecurve_list):
        
        lc_instance = cls()
        lc_instance = lc_instance._set_empty_attribute_arrays()

        for i, lc in enumerate(litecurve_list):
            for k in lc_instance.__dict__.keys():
                if type(lc_instance.__dict__[k]) is np.ndarray:
                    lc_instance.__setattr__(k, np.hstack([lc_instance.__dict__[k],lc.__dict__[k]]))

        return lc_instance
    

    @classmethod
    def from_kplr_pdcsap(cls, data_dir, target_id, obsmode, visits=None):
        """
        Load photometric data from Kepler Project PDCSAP Flux lightcurves
        The raw fits files must be pre-downloaded from MAST servers and stored locally
        
        This function performs minimal detrending steps
         * remove_nans()
         * normalize()
                
        Args:
            data_dir (str) : path to where data are stored
            target_id (int) : KIC number
            obsmode (str) : 'short cadence' or 'long cadence'
            visits (list) : optional, list of visits (Kepler quarters) to load.
        Returns:
            LiteCurve : self
        """

        # create instance of litecurve
        lc_instance = cls()
        lc_instance.mission = "Kepler"

        # sanitize inputs
        if visits is None:
            visits = np.arange(18, dtype=int) # hard coded for Kepler
        if isinstance(visits, int):
            visits = [visits]

        # load the raw MAST files using lightcurve
        mast_files = glob.glob(data_dir + f"kplr{target_id:09d}*.fits") # hard-coded for Kepler
        mast_files.sort()
        
        mast_data_list = []
        for i, mf in enumerate(mast_files):
            with fits.open(mf) as hdu_list:
                if hdu_list[0].header["OBSMODE"] == obsmode and np.isin(
                    hdu_list[0].header["QUARTER"], visits # hard coded for Kepler
                ):
                    mast_data_list.append(lk.read(mf))

        lk_col_raw = lk.LightCurveCollection(mast_data_list)

        # clean up the Collection data structure
        visits = []
        for lkc in lk_col_raw:
            visits.append(lkc.quarter) # hard coded for Kepler

        lk_col_clean = []
        for v in np.unique(visits):
            lkc_list = []
            cadno = []

            for lkc in lk_col_raw:
                if (lkc.quarter == v) * (lkc.targetid == target_id): # hard coded for kepler
                    lkc_list.append(lkc)
                    cadno.append(lkc.cadenceno.min())

            order = np.argsort(cadno)
            lkc_list = [lkc_list[j] for j in order]

            # lk.stitch() also normalizes the lightkurves
            lkc = lk.LightCurveCollection(lkc_list).stitch().remove_nans()
            
            lkc.quarter = lkc.quarter*np.ones(len(lkc.time), dtype='int') # hard coded for kepler
            lkc.season = lkc.quarter % 4 # hard coded for Kepler
            
            lk_col_clean.append(lkc)

        lk_col_clean = lk.LightCurveCollection(lk_col_clean)

        # stitch into a single LightCurve
        lklc = lk_col_clean.stitch()

        # set LiteCurve attributes
        lc_instance.time = np.array(lklc.time.value, dtype=float)
        lc_instance.flux = np.array(lklc.flux.value, dtype=float)
        lc_instance.error = np.array(lklc.flux_err.value, dtype=float)
        lc_instance.cadno = np.array(lklc.cadenceno.value, dtype=int)
        lc_instance.visit = np.array(lklc.quarter, dtype=int) # hard coded for Kepler
        lc_instance.obsmode = np.array([obsmode]*len(lc_instance.cadno), dtype=str)
        lc_instance.quality = np.array(lklc.quality.value, dtype=int)
        lc_instance.season = np.array(lklc.season, dtype=int)
        
        # remove cadences flagged by Kepler project pipeline
        lc_instance = lc_instance._remove_flagged_cadences(lklc.quality)

        return lc_instance
        

    @classmethod
    def from_k2(cls, data_dir, target_id, obsmode, visits=None):
        raise NotImplementedError("Loading K2 data not yet implemented")
    
    
    @classmethod
    def from_tess(cls, data_dir, target_id, obsmode, visits=None):
        raise NotImplementedError("Loading TESS data not yet implemented")

        
    @classmethod
    def from_alderaan(cls, data_dir, target_id):
        raise NotImplementedError("Loading ALDERAAN files not yet implemented")
    
    
    def split_visits(self):
        visits = np.unique(self.visit)

        litecurve_list = []
        for v in visits:
            litecurve = LiteCurve()
            for k in self.__dict__.keys():
            # for k in litecurve.__dict__.keys():
                # if type(litecurve.__dict__[k]) is np.ndarray:
                if type(self.__dict__[k]) is np.ndarray:
                    litecurve.__setattr__(k, self.__dict__[k][self.visit == v])
            litecurve_list.append(litecurve)

        return litecurve_list
      

    def _remove_flagged_cadences(self, quality_flags, bitmask='default'):
        qmask = lk.KeplerQualityFlags.create_quality_mask(
            quality_flags, bitmask=bitmask
        )
        for k in self.__dict__.keys():
            if type(self.__dict__[k]) is np.ndarray:
                self.__setattr__(k, self.__dict__[k][qmask])

        self.quality = np.ones(len(self.time), dtype=bool)

        return self



__all__ = ['QualityControl']

from astropy.stats import mad_std
import numpy as np
from alderaan.constants import kepler_lcit, kepler_scit
from alderaan.modules.base import BaseAlg
import warnings


class QualityControl(BaseAlg):
    def __init__(self, litecurve, planets):
        super().__init__(litecurve, planets)


    def check_coverage(self):
        """
        Check that each transit has at least 50% coverage
          * number of cadences in transit (tc +/- 0.5*T14)
          * number of cadences near transit (tc +/- 1.5*T14)
        """
        lc = self.litecurve

        quality = [None]*self.npl
        obsmode = self.get_transit_obsmode()

        for n, p in enumerate(self.planets):
            exptime = np.array([self._obsmode_to_exptime(om) for om in obsmode[n]])
            count_expected = np.maximum(1, np.array(p.duration / exptime, dtype=int))

            quality[n] = np.zeros(len(p.ephemeris.ttime), dtype='bool')

            for i, tc in enumerate(p.ephemeris.ttime):

                in_transit = np.abs(lc.time - tc) / p.duration < 0.5
                near_transit = np.abs(lc.time - tc) / p.duration < 1.5

                enough_pts_in = np.sum(in_transit) >= 0.5 * count_expected[i]
                enough_pts_near = np.sum(near_transit) >= 1.5 * count_expected[i]

                quality[n][i] = enough_pts_in and enough_pts_near

            f_quality = np.sum(quality[n]) / len(quality[n])

            if f_quality < 0.8:
                warnings.warn(f"WARNING: {int((1-f_quality)*100)}% of transits for Planet {n} have been flagged as low quality")

        return quality

            
    def check_rms(self, rel_size=None, abs_size=None, sigma_cut=5.0):
        if (rel_size is None) and (abs_size is None):
            raise ValueError("either rel_size or abs_size must be provided")
        
        if rel_size is None:
            rel_size = 0.
        if abs_size is None:
            abs_size = 0.

        lc = self.litecurve

        quality = [None]*self.npl
        overlap = self.identify_overlapping_transits(rtol=rel_size, atol=abs_size)
        obsmode = self.get_transit_obsmode()

        for n, p in enumerate(self.planets):
            rms = np.zeros(len(p.ephemeris.ttime))
            half_width = 0.5*np.max([rel_size*p.duration, abs_size])

            for i, tc in enumerate(p.ephemeris.ttime):
                flux = lc.flux[np.abs(lc.time - tc) < half_width]
                rms[i] = mad_std(flux, ignore_nan=True)

            quality[n] = np.ones(len(p.ephemeris.ttime), dtype='bool')

            for om in np.unique(np.array(obsmode[n])):
                use = (obsmode[n]== om) & overlap[n]
                rms_mu = np.nanmedian(rms[use])
                rms_sd = mad_std(rms[use], ignore_nan=True)

                quality[n][use] &= np.abs(rms[use] - rms_mu)/rms_sd < sigma_cut

        return quality


__all__ = ['plot_omc',
           'plot_litecurve',
           'dynesty_runplot',
           'dynesty_traceplot',
           'dynesty_cornerplot',
          ]

from astropy.stats import mad_std
from dynesty import plotting as dyplot
import matplotlib.pyplot as plt
import numpy as np
from alderaan.planet import Planet
from alderaan.ephemeris import Ephemeris
from alderaan.modules.omc import OMC


def plot_omc(data, target, filepath=None, interactive=False):
    """
    Plot observed-minus-calculated
    Input can be an alderaan Ephmeris or OMC object or a list of these
    """
    # check inputs
    if isinstance(data, OMC):
        omc_list = [data]
    elif isinstance(data, Ephemeris):
        omc_list = [OMC(data)]
    elif isinstance(data, list) & all([isinstance(d, OMC) for d in data]):
        omc_list = [d for d in data]
    elif isinstance(data, list) &  all([isinstance(d, Ephemeris) for d in data]):
        omc_list = [OMC(d) for d in data]
    else:
        raise ValueError("expected input: Ephemeris or OMC, or a list of these")
        
    npl = len(omc_list)

    # set colors
    if np.any([omc.out_prob is None for omc in omc_list]):
        marker_color = ['lightgrey']*npl
        line_color = [f'C{n}' for n in range(npl)]
        cmap = None
    else:
        marker_color = [1 - omc.out_prob for omc in omc_list]
        line_color = ['k' for n in range(npl)]
        cmap = 'viridis'
    
    # make plot
    fig, ax = plt.subplots(npl, figsize=(8, 3*npl))
    if npl == 1:
        ax = [ax]

    for n, omc in enumerate(omc_list):
        ax[n].scatter(omc.xtime, omc.yobs*24*60, c=marker_color[n], cmap=cmap, label='Observed TTVs')
        if omc.ymod is not None:
            ax[n].plot(omc.xtime, omc.ymod*24*60, lw=2, c=line_color[n], label='Regularized Model')
        if len(omc.quality) > 0:
            ax[n].plot(omc.xtime[~omc.quality], omc.yobs[~omc.quality]*24*60, 'rx')

        if omc.yerr is not None:
            err = np.nanmedian(omc.yerr) * 24 * 60
            ax[n].text(
                0.05,
                0.10,
                f"measured error = {err:.1f} min",
                transform=ax[n].transAxes,
                fontsize=12,
                ha='left',
                backgroundcolor='w',
            )
            
        if omc.ymod is not None:
            rms = mad_std(omc.yobs - omc.ymod[n], ignore_nan=True) * 24 * 60
            ax[n].text(
                0.95,
                0.10,
                f"residual RMS = {rms:.1f} min",
                transform=ax[n].transAxes,
                fontsize=12,
                ha='right',
                backgroundcolor='w',
            )

        ax[n].text(
            0.95,
            0.85,
            f"P = {omc._static_period:.1f}",
            transform=ax[n].transAxes,
            fontsize=14,
            ha='right',
            backgroundcolor='w'
        )

        ax[n].tick_params(labelsize=12)
        ax[n].set_ylabel("O-C [min]", fontsize=20)
        
        yrange = 24*60*np.max([np.abs(omc.yobs.min()),np.abs(omc.yobs.max())])
        yrange = np.max([5, 1.1*yrange])
        ax[n].set_ylim(-yrange, +yrange)

    ax[0].set_title(f"{target}", fontsize=20)
    ax[n].set_xlabel("Time [BJKD]", fontsize=20)

    plt.tight_layout()

    if filepath is not None:
        plt.savefig(filepath)
    if not interactive:
        plt.close(fig)
    
    return fig


def plot_litecurve(litecurve, target, planets=None, filepath=None, interactive=False):
    # shorthand
    lc = litecurve

    if planets is not None:
        if isinstance(planets, Planet):
            planets = [planets]
    
    fig, ax = plt.subplots(1,1, figsize=(20,4))
    ax.plot(lc.time, lc.flux, 'k.', ms=0.5)
    ax.tick_params(labelsize=12)
    ax.set_xlabel("Time [BJKD]", fontsize=24)
    ax.set_ylabel("Flux", fontsize=24)
    ax.set_xlim(lc.time.min(), lc.time.max())

    yrange = 1.3*np.max([
        np.abs(1-np.percentile(lc.flux, 0.1)),
        np.abs(1-np.percentile(lc.flux, 99.9))
    ])
    ax.set_ylim(1-yrange, 1+yrange)

    if planets is not None:
        ymin = 1 - yrange*(1.3 + 0.1*len(planets))
        ymax = 1 + yrange 
        ax.set_ylim(ymin, ymax)
        for n, p in enumerate(planets):
            ax.plot(
                p.ephemeris.ttime, 
                np.ones_like(p.ephemeris.ttime) - (0.9-0.1*n)*(1-ymin), 
                '^',
                c=f'C{n}'
            )

    plt.tight_layout()

    if filepath is not None:
        plt.savefig(filepath)
    if not interactive:
        plt.close(fig)
    
    return fig, ax


def dynesty_runplot(results, target, filepath=None, interactive=False):
    fig, ax = dyplot.runplot(results, logplot=True, label_kwargs={'fontsize':16}, color='#0d0887')
    
    ax[0].set_title(target, fontsize=24)
    fig.tight_layout()
    
    if filepath is not None:
        plt.savefig(filepath, bbox_inches='tight', dpi=96)
    if not interactive:
        plt.close(fig)

    return fig, ax


def dynesty_traceplot(results, target, planet_no, filepath=None, interactive=False):
    fig, ax = dyplot.traceplot(
        results,
        labels=_parameter_labels(1, subscripts=False),
        dims=np.arange(5 * planet_no, 5 * (planet_no + 1)),
        label_kwargs={'fontsize':14},
    )
    
    fig.tight_layout()
    fig.suptitle(f"{target} - Planet {planet_no}", fontsize=18, y=fig.subplotpars.top + 0.02)
    
    if filepath is not None:
        plt.savefig(filepath, bbox_inches='tight', dpi=96)
    if not interactive:
        plt.close(fig)
        
    return fig, ax


def dynesty_cornerplot(results, target, planet_no, filepath=None, interactive=False):
    fig, ax = dyplot.cornerplot(
        results,
        labels=_parameter_labels(1, subscripts=False),
        dims=np.arange(5 * planet_no, 5 * (planet_no + 1)),
        label_kwargs={'fontsize':14},
        color=f'C{planet_no}',
    )
    
    fig.tight_layout()
    fig.suptitle(f"{target} - Planet {planet_no}", fontsize=18, y=fig.subplotpars.top + 0.02)
    
    if filepath is not None:
        plt.savefig(filepath, bbox_inches='tight', dpi=96)
    if not interactive:
        plt.close(fig)
        
    return fig, ax


def _parameter_labels(npl, subscripts=True):
    labels = []
    
    if subscripts:
        for n in range(npl):
            labels = labels + f'$C0_{n}$ $C1_{n}$ $r_{n}$ $b_{n}$ $T14_{n}$'.split()
    else:
        for n in range(npl):
            labels = labels + f'C0 C1 r b T14'.split()

    labels += 'q1 q2'.split()
    
    return labels